/*
    Soundblaster 16 PCI driver for Solaris.
    @(#) sbpci.h 2.1@(#)

    Written by Philip Brown <phil@bolthole.com>
    Rewritten by Jrgen Keil <jk@tools.de>

    Copyright  2000      Philip Brown
    Copyright  2000-2002 Jrgen Keil

    This driver is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This driver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this driver; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
*/

#if	_KERNEL

#include <sys/sunddi.h>
#include <sys/ksynch.h>
#include <sys/note.h>
#include <sys/mixer.h>
#include <sys/audio.h>
#include <sys/audio/audio_helper.h>
#include "audio_codec.h"

/*
 * hide gcc2 __attribute__ compiler extension from non-gcc compiler
 * and old gcc compiler
 */
#if     !defined(__GNUC__)
#define __attribute__(x)        /**/
#elif   __GNUC__ < 2
#define __attribute__(x)        /**/
#endif


/*
 * Device name returned in the name field of the audio_device
 * structure in the AUDIO_GETDEV ioctl.
 *
 * openwin "audiotool" knows about these HW names and derives
 * valid playback formats from them. Uhg.
 * For this reason the string returned in the audio_device.name
 * field is configurable with a sbpci.conf string property.
 */
#ifndef	DEFAULT_DEVICE_NAME
#define	DEFAULT_DEVICE_NAME	"TOOLS,sbpci"
#endif


/* 
 * The ES1371 has two DACs (1/2) for playback.  Choose which
 * one the driver uses. 
 *
 * DAC1 has a special "sample count reload" bit in the serial
 * interface register to immediately notify the hardware of changes
 * for the sample counter. This allows precise timing when the
 * hardware sends the next playback interrupt.  This feature is not
 * available for DAC2.
 *
 * DAC2 can playback the last sample (instead of zero samples)
 * when DAC2 is stopped.  This avoids clicks when we reload the
 * DMA counters.  This feature is not available for DAC1.
 *
 * Previous versions of the driver used DAC2 for the sbpci driver
 * because the DMA counters were reloaded every 8KBytes.  Now that we
 * do not reload the DMA counter any more during playback we're using
 * DAC1 instead.
 *
 * (For unknown reasons, DAC1 can be stopped "immediately", but
 * there's a small delay before a disable on DAC2 is effective. OTOH
 * DAC2 has an advantage on the ENS1370 because it supports variable
 * samplerates via the PLL clock generator)
 */
#ifndef	DEFAULT_PLAYBACK_DAC
#define	DEFAULT_PLAYBACK_DAC	(softstate->playback_dac)
#endif


/*
 * Default "gain" value after loading the driver
 */
#ifndef	DEFAULT_PLAY_GAIN
#define	DEFAULT_PLAY_GAIN	AUDIO_MID_GAIN
#endif
#ifndef	DEFAULT_RECORD_GAIN
#define	DEFAULT_RECORD_GAIN	AUDIO_MID_GAIN
#endif


/*
 * sample buffer with data in the device's native format.
 * The sample buffer is allocated in DMAable memory.
 *
 * This buffer is used as a ring buffer.  It contains 'used' bytes of
 * valid sample data starting at byte offset 'offs'.  The ring buffer
 * has a total capacity of SAMPLEBUF_SIZE bytes, and all offset
 * calculations are performed "mod SAMPLEBUF_SIZE".
 *
 * For playback, the sound card hardware accesses the sample data ring
 * buffer in chunks, starting at offset 'offs' upto 'offs+used' - but
 * not more than SAMPLEBUF_INTR_CHUNK bytes at one time.  When
 * playback of a chunk is finished, the sound cards sends an
 * interrupt and the driver advances 'offs' and decreases 'used', thus
 * freeing the chunk just played.
 *
 * New samples are filled into the buffer at offset 'offs+used'.
 */
#define	SAMPLEBUF_INTR_CHUNK	0x2000		/* one intr every 8KByte */
#define	SAMPLEBUF_SIZE		(3*SAMPLEBUF_INTR_CHUNK)


typedef enum sample_buf_state {
    SAMPLE_BUF_IDLE	= 0,		/* not playing */
    SAMPLE_BUF_STREAMING_AVAIL,		/* DMA active, buffer space avail. */
    SAMPLE_BUF_STREAMING_FULL		/* DMA active, no space available */
} sample_buf_state_t;


typedef struct sample_buffer {
    sample_buf_state_t	state;
    caddr_t		data;		/* kernel virt addr for samples */
    uint32_t		offs;		/* byte offset into data buffer where current DMA started */
    uint32_t		used;		/* space used starting at offs */
    ddi_dma_handle_t	dma_handle;
    uint8_t		dma_bound;
    size_t	    	reallen;	/* real length of DMA buffer, */
					/* includes alignments, ... */
    ddi_acc_handle_t	acc_handle;
} sample_buffer_t;



typedef struct ens_chip_ops ens_chip_ops_t;


typedef struct audio1371_softstate {
    dev_info_t         *dip;            /* device info node for your sound hw */
    audiohdl_t		audio_handle;	/* opaque handle to audio support module */

    kmutex_t            instance_mutex;
    kmutex_t            intr_mutex;
    ddi_iblock_cookie_t	intr_cookie;

    caddr_t		regs;
    ddi_acc_handle_t	regs_handle;
    ddi_acc_handle_t	pci_conf_handle;
    
    int                 intr_count;     /* debug info */

    ens_chip_ops_t     *ens_chip_ops;	/* audio controller dependant ops */
    audio_codec_handle_t codec_handle;	/* AC97 or AK4531 codec "object" */

    char               *revstr;         /* pci revision of card as printable string */
    uint8_t             rev;            /* pci revision byte of card */
    uint8_t             ens_chip_type;  /* ENS_CHIP_1370, ENS_CHIP_1371, ... */
    uint8_t             ct5880;         /* TRUE: using CT5880 chip */
    uint8_t		playback_dac;	/* 1/2: DAC to use for playback */

    sample_buffer_t     record_buf;     /* buffer space for ADC samples */
    sample_buffer_t     play_buf;       /* sample data for the DAC playback */
    timeout_id_t	play_timeout_id;

    audio_info_t	audio_info;	/* driver local audio state. Note: */
    					/* we maintain only a few fields */
    					/* of the structure in the driver */

    audio_device_t	dev_info;
    ah_ad_info_t	ad_info;
} audio1371_softstate_t;

/* possible values for ens_chip_type field */
#define	ENS_CHIP_1370 0
#define ENS_CHIP_1371 1


/* 1370/1371 chip-specific subroutines */
struct ens_chip_ops {
    audio_codec_handle_t (*codec_init)(struct audio1371_softstate *);

    void (*codec_interface_reset)(struct audio1371_softstate *softstate);

    int (*configure_pm)(struct audio1371_softstate *softstate, int level);
    int (*init_sample_rate_conv)(struct audio1371_softstate *softstate);
    int (*enable_sample_rate_conv)(struct audio1371_softstate *softstate,
				   int dacnum,
				   int enable_flag,
				   int *old_enable_flag);

    void (*setADCfreq)(struct audio1371_softstate *softstate, int rate);
    void (*setDACfreq)(struct audio1371_softstate *softstate, int rate, int dacnum);
    
};

extern ens_chip_ops_t ens1370_ops, ens1371_ops;


#define	ENS_CODEC_INIT(state)						\
		(state)->ens_chip_ops->codec_init((state))
#define	ENS_CODEC_INTERFACE_RESET(state)				\
		(state)->ens_chip_ops->codec_interface_reset((state))
#define	ENS_CONFIGURE_PM(state, level)					\
		(state)->ens_chip_ops->configure_pm((state), (level))

#define	ENS_INIT_SAMPLE_RATE_CONV(state)				\
		(state)->ens_chip_ops->init_sample_rate_conv((state))
#define	ENS_ENABLE_SAMPLE_RATE_CONV(state, dac, flag, old_flag)		\
		(state)->ens_chip_ops->enable_sample_rate_conv(		\
				(state), (dac), (flag), (old_flag))
#define	ENS_SETADCFREQ(state, rate)					\
		(state)->ens_chip_ops->setADCfreq((state), (rate))
#define	ENS_SETDACFREQ(state, rate, dac)				\
		(state)->ens_chip_ops->setDACfreq((state), (rate), (dac))


#define	TRUE	1
#define	FALSE	0

/* general success/failure return codes */
#define	AUDIO1371_SUCCESS	1
#define	AUDIO1371_FAILURE	2



/* available log levels */
#define	AUD1371_LOG_L0	0	/* warnings, console & syslog buffer */
#define	AUD1371_LOG_L1	1	/* errors, syslog buffer */
#define	AUD1371_LOG_L2	2	/* recoverable errors, debug only */
#define	AUD1371_LOG_L3	3	/* interesting data, debug only */
#define	AUD1371_LOG_L4	4	/* tracing, debug only */


/* available log masks */
#define	AUD1371M_AC97	0x00000020	/* AC97 register access */
#define	AUD1371M_CODEC	0x00000010	/* CODEC IO */
#define	AUD1371M_INTR	0x00000008	/* interrupt handling */
#define	AUD1371M_DRIVER	0x00000004	/* audio driver functions */
#define	AUD1371M_ATTACH	0x00000002	/* card setup at attach time */
#define	AUD1371M_MODS	0x00000001	/* _init, _info, _fini */
#define	AUD1371M_ALL	(~0u)


/* logging for log levels 4..0 */
#if	DEBUG
#define	DPRINTF4(...)	audio1371_dprintf4(__VA_ARGS__)
#define	DPRINTF3(...)	audio1371_dprintf3(__VA_ARGS__)
#define	DPRINTF2(...)	audio1371_dprintf2(__VA_ARGS__)
#else
#define	DPRINTF4(...)	/**/
#define	DPRINTF3(...)	/**/
#define	DPRINTF2(...)	/**/
#endif
#define	DPRINTF1(...)	audio1371_dprintf1(__VA_ARGS__)
#define	DPRINTF0(...)	audio1371_dprintf0(__VA_ARGS__)


#if	DEBUG
extern void __attribute__ ((format (__printf__, 2, 3)))
     audio1371_dprintf4(uint_t mask, char *fmt, ...);
extern void __attribute__ ((format (__printf__, 2, 3)))
     audio1371_dprintf3(uint_t mask, char *fmt, ...);
extern void __attribute__ ((format (__printf__, 2, 3)))
     audio1371_dprintf2(uint_t mask, char *fmt, ...);
#endif
extern void __attribute__ ((format (__printf__, 2, 3)))
     audio1371_dprintf1(uint_t mask, char *fmt, ...);
extern void __attribute__ ((format (__printf__, 2, 3)))
     audio1371_dprintf0(uint_t mask, char *fmt, ...);



/*
 * shortcuts for reading/writing the sbpci card's i/o registers
 */
#define	SBPCI_GET8(offs)						\
	ddi_get8 (softstate->regs_handle,				\
 		  (void*)((char*)softstate->regs + (offs)))
#define	SBPCI_GET16(offs)						\
	ddi_get16(softstate->regs_handle,				\
		  (void*)((char*)softstate->regs + (offs)))
#define	SBPCI_GET32(offs)						\
	ddi_get32(softstate->regs_handle,				\
		  (void*)((char*)softstate->regs + (offs)))

#define	SBPCI_PUT8(offs, val)						\
	ddi_put8 (softstate->regs_handle,				\
		  (void*)((char*)softstate->regs + (offs)), (val))
#define	SBPCI_PUT16(offs, val)						\
	ddi_put16(softstate->regs_handle,				\
		  (void*)((char*)softstate->regs + (offs)), (val))
#define	SBPCI_PUT32(offs, val)						\
	ddi_put32(softstate->regs_handle,				\
		  (void*)((char*)softstate->regs + (offs)), (val))

#endif  /* _KERNEL */

/******************************************************************/
/* Here are the actual Ensoniq PCI register (byte) offsets. 	  */
/******************************************************************/

#define SBPCI_CONTROL		0x00	/* 4byte */
#define SBPCI_STATUS		0x04	/* 4byte */
#define SBPCI_UART_DATA		0x08	/* 1byte *//* UART is for MIDI */
 /* SBPCI_MEMPAGE defines which "page" the 16 bytes at SCPCI_MEMORY map to */
 /* Valid range is only 0x00-0x0f  */
#define SBPCI_MEMPAGE		0x0c	/* 4byte */

#define	SBPCI_1370_CODEC	0x10	/* 4byte */

#define	SBPCI_1371_SAMPLERATE	0x10	/* 4byte */

  /* codec control. This is actually an old */
  /* "port" style access. Ugh!              */
#define SBPCI_1371_CODEC	0x14	/* 4byte */

  /* This is only for legacy ENSONIQ it seems*/
#define SBPCI_LEGACY		0x18	/* 8byte */

#define SBPCI_SERIAL		0x20	/* 16byte */
#define	 SBPCI_DAC1_COUNT	0x24	/* 2*2byte sample counter */
#define	 SBPCI_DAC2_COUNT	0x28	/* 2*2byte sample counter */
#define	 SBPCI_ADC_COUNT	0x2c	/* 2*2byte sample counter */


#define SBPCI_MEMORY		0x30	/* 16byte */

  /* memory page 0x0c: */
#define  SBPCI_DAC1_FRAME1	0x30
#define  SBPCI_DAC1_FRAME2	0x34
#define  SBPCI_DAC2_FRAME1	0x38
#define  SBPCI_DAC2_FRAME2	0x3c
  /* memory page 0x0d: */
#define  SBPCI_ADC_FRAME1	0x30
#define  SBPCI_ADC_FRAME2	0x34
  /* memory page 0x0e or 0x0f: */
#define  SBPCI_UART_FIFO	0x30

#define	SBPCI_REG_SIZE		0x40	/* size of register file for the sbpci card */


  /* sample cache memory pages mapped to register 0x30 .. 0x3f */
#define	 SBPCI_DAC1_SAMP_MEMPG	0x00	/* 4 memory pages for DAC1 samples */
#define	 SBPCI_DAC2_SAMP_MEMPG	0x04	/* 4 memory pages for DAC2 samples */
#define	 SBPCI_ADC_SAMP_MEMPG	0x08	/* 4 memory pages for ADC samples */
#define	 SBPCI_DAC_MEMPG	0x0c	/* memory page for DAC12 frame information */
#define	 SBPCI_ADC_MEMPG	0x0d	/* memory page for ADC frame information */
#define	 SBPCI_UART_MEMPG	0x0e	/* 2 memory page for UART fifo */


/* Control bits for use with SBPCI_CONTROL register (0x00) */
#define SBPCI_CTRL_GPIO_INMASK	0x00f00000
#define SBPCI_CTRL_GPIO_OUTMASK	0x000f0000
#define SBPCI_CTRL_MSFMTSEL	0x00008000
#define SBPCI_CTRL_AC97WARMRESET 0x00004000	/* SYNC_RES */
#define SBPCI_CTRL_ADC_STOP	0x00002000
#define SBPCI_CTRL_PWR_INTRM	0x00001000
#define SBPCI_CTRL_M_CB		0x00000800
#define SBPCI_CTRL_CCB_INTRM	0x00000400
#define SBPCI_CTRL_PDLEV_MASK	0x00000300	/* current power down level D0-D3 */
#define	SBPCI_CTRL_PDLEV_SHIFT	8
/* These enable the specific resource */
#define SBPCI_CTRL_BREQ		0x00000080
#define SBPCI_CTRL_DAC1		0x00000040
#define SBPCI_CTRL_DAC2		0x00000020
#define SBPCI_CTRL_ADC		0x00000010
#define SBPCI_CTRL_UART		0x00000008
#define SBPCI_CTRL_JOYSTICK	0x00000004
/* obviously, these stop "clocks" */
#define SBPCI_CTRL_STOPXTL	0x00000002
#define SBPCI_CTRL_STOPCLK	0x00000001	/* stop ALL clocks */

#define	SBPCI_1370_CTRL_PCLKDIV_MASK 0x1fff0000	/* ens1370: DAC2 PLL */
#define	SBPCI_1370_CTRL_PCLKDIV_SHIFT 16
#define	SBPCI_1370_CTRL_WTSRSEL	0x00003000	/* ens1370: DAC1 fixed freq */
#define	SBPCI_1370_CTRL_5512	0x00000000
#define	SBPCI_1370_CTRL_11025	0x00001000
#define	SBPCI_1370_CTRL_22050	0x00002000
#define	SBPCI_1370_CTRL_44100	0x00003000
#define SBPCI_1370_CTRL_CODEC	0x00000002	/* ens1370: codec enable */
#define SBPCI_1370_CTRL_SERR_DIS 0x00000001	/* ens1370: serr disable */

#define SBPCI_1370_CLK2		1411200
#define SBPCI_1370_SR2PCLKDIV(x) (SBPCI_1370_CLK2/(x) - 2)

/*
 * STATUS bit flags for SBPCI_STATUS register (0x04). 
 * Mostly just for "Has interrupt occurred?"
 */
#define SBPCI_STAT_INTR		0x80000000	/* interrupt has happened*/
#define SBPCI_STAT_ALWAYS1	0x4ffffe00

#define SBPCI_STAT_CT5880_AC97_RESET 0x20000000	/* */

#define	SBPCI_1370_STAT_CSTAT	0x00000400	/* codec reg access status */
#define	SBPCI_1370_STAT_CBUSY	0x00000200	/* ENS 1370 only */
#define	SBPCI_1370_STAT_CWRIP	0x00000100

#define SBPCI_1371_STAT_SYNCERR	0x00000100	/* codec sync err */

/*                             0x08 & 0x04 == "voice code for CCB" */
#define SBPCI_1371_STAT_PWR	0x00000020	/* power interrupt(?) */
#define SBPCI_STAT_CCB		0x00000010	/* CCB interrupt */
#define SBPCI_STAT_UART		0x00000008	/* UART interrupt */
#define SBPCI_STAT_DAC1		0x00000004	/* DAC1 interrupt */
#define SBPCI_STAT_DAC2		0x00000002	/* DAC2 interrupt */
#define SBPCI_STAT_ADC		0x00000001	/* ADC interrupt */

/* Defines for use with the "samplerate" register*/
#define SBPCI_SAMPLE_ADDR	0xfe000000	/* SRC_RAM_ADR */
#define SBPCI_SAMPLE_WR_EN	0x01000000	/* write enable: SRC_RAM_WE*/
#define SBPCI_SAMPLE_BUSY	0x00800000	/* SRC_RAM_BUSY */
#define SBPCI_SAMPLE_DISABLE	0x00400000	/* disable ALL: SRC_DISABLE */
#define SBPCI_SAMPLE_DIS_DAC1	0x00200000	/* DIS_P1 */
#define SBPCI_SAMPLE_DIS_DAC2	0x00100000	/* DIS_P2 */
#define SBPCI_SAMPLE_DIS_ADC	0x00080000	/* DIS_REC */
#define SBPCI_SAMPLE_DISMASK	0x00780000	/* all the disablemasks*/

#define SBPCI_SAMPLE_CODECMAGIC	0x00010000	/* undocumented ALSA thing*/
#define SBPCI_SAMPLE_VALUE	0x0000ffff	/* the bits that will be
						 * transfered to internal regs
						 */

/* These are "addresses" through the SAMPLE RATE CONVERTER, above */
/* Each one gives access to a uint16 */
#define ENS_SAMPLE_ADC_VOL	0x6c
#define ENS_SAMPLE_ADC_VOL2	0x6d	/* if we had stereo recording */

#define ENS_SAMPLE_DAC1_TRUNC	0x70
#define ENS_SAMPLE_DAC1_INT	0x71
#define ENS_SAMPLE_DAC1_ACCUM	0x72	/* not used */
#define ENS_SAMPLE_DAC1_VFREQ	0x73
#define ENS_SAMPLE_DAC2_TRUNC	0x74
#define ENS_SAMPLE_DAC2_INT	0x75
#define ENS_SAMPLE_DAC2_ACCUM	0x76	/* not used */
#define ENS_SAMPLE_DAC2_VFREQ	0x77
#define ENS_SAMPLE_ADC_TRUNC	0x78
#define ENS_SAMPLE_ADC_INT	0x79
#define ENS_SAMPLE_ADC_ACCUM	0x7a	/* not used */
#define ENS_SAMPLE_ADC_VFREQ	0x7b

#define ENS_SAMPLE_DAC1_VOL	0x7c	/*left*/
#define ENS_SAMPLE_DAC1_VOL2	0x7d	/*right*/
#define ENS_SAMPLE_DAC2_VOL	0x7e	/*left*/
#define ENS_SAMPLE_DAC2_VOL2	0x7f	/*right*/


#define ENS_SAMPLE_SET_TRUNC(n)		((n)<<9)
#define ENS_SAMPLE_SET_N(n)		((n)<<4)
#define ENS_SAMPLE_SMF			0x8000
#define ENS_SAMPLE_SET_VFI(n)		((n)<<10)
#define ENS_SAMPLE_SET_ACI(n)		(n)
#define ENS_SAMPLE_SET_ADC_VOL(n)	((n)<<8)
#define ENS_SAMPLE_SET_DAC_VOLI(n)	((n)<<12)
#define ENS_SAMPLE_SET_DAC_VOLF(n)	(n)


#define	SBPCI_SERIAL_DAC_TEST	0x00400000
#define	SBPCI_SERIAL_P2_END_INC_MASK	0x00380000
#define	SBPCI_SERIAL_P2_ST_INC_MASK	0x00070000
#define	SBPCI_SERIAL_P2_END_INC(i)	((i) << 19)
#define	SBPCI_SERIAL_P2_ST_INC(i)	((i) << 16)

/* These enable INTERRUPTS for the specific resource, and are also the
 * mask for clearing interrupt status
 * To clear, zero the bit, then set it to re-enable another trigger.
 */
#define SBPCI_SERIAL_NOLOOPADC	0x00008000
#define SBPCI_SERIAL_NOLOOPDAC2	0x00004000
#define SBPCI_SERIAL_NOLOOPDAC1	0x00002000
#define SBPCI_SERIAL_PAUSEDAC2	0x00001000
#define SBPCI_SERIAL_PAUSEDAC1	0x00000800

#define SBPCI_SERIAL_ENABLEADC	0x00000400
#define SBPCI_SERIAL_ENABLEDAC2	0x00000200
#define SBPCI_SERIAL_ENABLEDAC1	0x00000100

#define SBPCI_SERIAL_SCT_RLD_DAC1 0x00000080	/* yes only for DAC1*/
#define	SBPCI_SERIAL_DAC2_SEN	0x00000040	/* play last sample in stop mode */

/* These change 8/16bitsound, and stereo/mono */ 
#define SBPCI_SERIAL_ADCW16	0x00000020
#define SBPCI_SERIAL_ADCSTEREO	0x00000010
#define SBPCI_SERIAL_DAC2W16	0x00000008
#define SBPCI_SERIAL_DAC2STEREO	0x00000004
#define SBPCI_SERIAL_DAC1W16	0x00000002
#define SBPCI_SERIAL_DAC1STEREO	0x00000001


/* When accessing the codec, through the "codec interface" register */
#define SBPCI_CODEC_READBIT	0x00800000
#define SBPCI_CODEC_READBSY	0x40000000
#define SBPCI_CODEC_READRDY	0x80000000 /* data from read-request ready */

/* pci revision -> chip mapping */
#define SBPCI_EV1938_A  0x00
/* #define SBPCI_CT5770?_A  0x01 */
#define SBPCI_CT5880_C  0x02
#define SBPCI_CT5880_D  0x03
#define SBPCI_ES1373_A  0x04
#define SBPCI_ES1373_B  0x06
#define SBPCI_CT5880_A  0x07
#define SBPCI_ES1373_8  0x08
#define SBPCI_ES1371_B  0x09

/* actual AC97 Codec read/write defines */
/* #include "ac97.h" */
