/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Intel technical specifications for the ICH audio controller:
 *
 * Intel 82801AA (ICH) & Intel 82801AB (ICH0) I/O Controller Hub
 * AC'97 Programmer's Reference Manual
 * (Order Number: 298028-001)
 *
 * Intel 82801AA (ICH) and Intel 82801AB (ICH0) I/O Controller Hub
 * (Order Number: 290655-003)
 *
 * Intel 82801DB I/O Controller Hub 4 (ICH4)
 * (Document Number: 290744-001)
 */


#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#if	SUNOS >= 509
#include <sys/audio/ac97.h>
#else
#include <sys/ac97.h>
#endif
#include <sys/audio.h>
#include "audioio_compat.h"
#include "ac97_codec.h"
#include "audioi810.h"

#include <sys/ddi.h>

/*
 * hide gcc2 __attribute__ compiler extension from non-gcc compiler
 * and old gcc compiler
 */
#if     !defined(__GNUC__)
#define __attribute__(x)        /**/
#elif   __GNUC__ < 2
#define __attribute__(x)        /**/
#endif



static int audioi810_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int audioi810_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int audioi810_power(dev_info_t *dip, int component, int level);
static int audioi810_fill_play_buffers(struct audioi810_softstate *softstate);
static int audioi810_prepare_record_buffers(struct audioi810_softstate *softstate);

static audio_info_t audioi810_defaults;



static struct module_info audioi810_minfo = {
    0x810,		/* module ID number */
    "audi810",		/* module name */
    0,			/* minimum packet size accepted */
    /*4096*/8192,		/* maximum packet size accepted */
    48000*4*2,		/* hi-water mark: ~2 secs of audio */
    0x8000		/* lo-water mark */
};


static struct qinit audioi810_rqueue = {
    audio_hlp_rput,
    audio_hlp_rsvc,
    audio_hlp_open,
    audio_hlp_close,
    NULL,
    &audioi810_minfo
};

static struct qinit audioi810_wqueue = {
    audio_hlp_wput,
    audio_hlp_wsvc,
    NULL,
    NULL,
    NULL,
    &audioi810_minfo
};


static struct streamtab audioi810_str_info = {
    &audioi810_rqueue,
    &audioi810_wqueue,
    NULL,
    NULL,
};


/*
 * char/block ops vector for a STREAMS driver
 */
static struct cb_ops audioi810_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &audioi810_str_info,	/* cb_str */
    D_NEW|D_MTSAFE|D_64BIT,	/* cb_flag */
    CB_REV,			/* cb_rev */
    nodev,			/* cb_aread */
    nodev,			/* cb_awrite */
};

static struct dev_ops  audioi810_drv_ops = {
    DEVO_REV,
    0,

    audio_hlp_getinfo,
    nulldev,			/* devo_identify */
    nulldev,			/* devo_probe */
    audioi810_attach,
    audioi810_detach,
    nodev,			/* devo_reset */

    &audioi810_cb_ops,
    NULL,			/* devo_bus_ops */
    audioi810_power,		/* devo_power */
};

/*
 * Module linkage information for the kernel.
 */
static struct modldrv audioi810_modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    "i810 audio driver",
    &audioi810_drv_ops		/* driver ops */
};


static struct modlinkage audioi810_modlinkage = {
    MODREV_1,
    { &audioi810_modldrv, NULL },
};


/*
 * This describes the ICH's DMA engines to the OS at hardware level
 * 
 * The first structure is for the "buffer descriptor list".
 * These must be in contiguous physical memory.
 *
 * The ICH accesses bdl memory in 8 bytes quantities (=> minxfer == 8)
 * and the whole bdl list has a fixed size of 32*8 == 256 bytes. A bdl list
 * must be allocated in contiguous physical memory (=> s/g length == 1).
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t bdl_dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x0000ffff,			/* DMA counter register */
    8,				/* DMA address alignment */
    				/*   bdl list must be aligned on a 8-byte boundary */
    0x3c,			/* DMA burstsizes ?? */
    8,				/* minxfer effective DMA size */
    0x0000ffff,			/* max DMA xfer size */
    0x00000fff,			/* segment boundary*/
    1,				/* s/g length */
    8,				/* granularity of device */
				/*   size of a bdl entry?? */
    0				/* flags */
};

/*
 * DMA attributes for sample data.  No need for physical contiguous
 * memory.
 *
 * The ICH accesses sample data in longword (4 bytes) quantities
 * (=> minxfer == 4) and has a 16-bit word DMA counter (=> max DMA
 * xfer size == 0xffff * 2).  It does support scatter/gatter DMA
 * transfers from upto 32 memory locations (=> s/g length == 32).
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x0001fffe,			/* DMA counter register */
    2,				/* DMA address alignment */
    0x3c,			/* DMA burstsizes */
    4,				/* minxfer effective DMA size */
    0x0001ffff,			/* max DMA xfer size */
    0x00001fff,			/* segment boundary*/
    I810_NUM_BD,		/* s/g length */
    4,				/* granularity of device */
				/*   (ICH dma operates on longwords) */
    0				/* flags */
};

/* This describes things like byteorder internally for the ICH */
static struct ddi_device_acc_attr dev_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_NEVERSWAP_ACC,
    DDI_STRICTORDER_ACC
};


/* available log levels */
#define	AUDI810_LOG_L0	0		/* warnings, console & syslog buffer */
#define	AUDI810_LOG_L1	1		/* errors, syslog buffer */
#define	AUDI810_LOG_L2	2		/* recoverable errors, debug only */
#define	AUDI810_LOG_L3	3		/* interesting data, debug only */
#define	AUDI810_LOG_L4	4		/* tracing, debug only */


/* available log masks */
#define	AUDI810M_AC97	0x00000020	/* AC97 register access */
#define	AUDI810M_CODEC	0x00000010	/* CODEC IO */
#define	AUDI810M_INTR	0x00000008	/* interrupt handling */
#define	AUDI810M_DRIVER	0x00000004	/* audio driver functions */
#define	AUDI810M_ATTACH	0x00000002	/* card setup at attach time */
#define	AUDI810M_MODS	0x00000001	/* _init, _info, _fini */
#define	AUDI810M_ALL	(~0u)

uint_t audioi810_errlevel = AUDI810_LOG_L2;
uint_t audioi810_errmask =  AUDI810M_ALL;


/* logging for log levels 4..0 */
#if	DEBUG
#define	DPRINTF4(...)	audioi810_dprintf4(__VA_ARGS__)
#define	DPRINTF3(...)	audioi810_dprintf3(__VA_ARGS__)
#define	DPRINTF2(...)	audioi810_dprintf2(__VA_ARGS__)
#else
#define	DPRINTF4(...)	/**/
#define	DPRINTF3(...)	/**/
#define	DPRINTF2(...)	/**/
#endif
#define	DPRINTF1(...)	audioi810_dprintf1(__VA_ARGS__)
#define	DPRINTF0(...)	audioi810_dprintf0(__VA_ARGS__)


static void
audioi810_vprintf(uint_t level, uint_t mask, char *fmt, va_list args)
{
    if (level <= audioi810_errlevel && (mask & audioi810_errmask))
	vcmn_err(CE_CONT, fmt, args);
}


#ifdef	DEBUG
void __attribute__ ((format (__printf__, 2, 3)))
audioi810_dprintf4(uint_t mask, char *fmt, ...)

{
    va_list args;
    va_start(args, fmt);
    audioi810_vprintf(AUDI810_LOG_L4, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audioi810_dprintf3(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioi810_vprintf(AUDI810_LOG_L3, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audioi810_dprintf2(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioi810_vprintf(AUDI810_LOG_L2, mask, fmt, args);
    va_end(args);
}
#endif

void __attribute__ ((format (__printf__, 2, 3)))
audioi810_dprintf1(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioi810_vprintf(AUDI810_LOG_L1, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audioi810_dprintf0(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioi810_vprintf(AUDI810_LOG_L0, mask, fmt, args);
    va_end(args);
}


/* 
 * Global softstate for multiple i810 audio controller instances
 */
static void *audioi810_state;


/* 
 * Sun driver module initialization and finalization code 
 */
int 
_init(void)
{
    int err;

    err = ddi_soft_state_init(&audioi810_state,
			      sizeof(struct audioi810_softstate), 1);
    if (err == 0) {
	if ((err = mod_install(&audioi810_modlinkage)) != 0)
	    ddi_soft_state_fini(&audioi810_state);
    }

    DPRINTF4(AUDI810M_MODS, "_init, err=%d\n", err);

    return err;
}


int
_fini(void)
{
    int err;
    
    if ((err = mod_remove(&audioi810_modlinkage)) == 0)
	ddi_soft_state_fini(&audioi810_state);

    DPRINTF4(AUDI810M_MODS, "_fini, err=%d\n", err);

    return err;
}


int
_info(struct modinfo *modinfop)
{
    return mod_info(&audioi810_modlinkage, modinfop);
}


/*
 * Allocate a list of DMA memory buffers (for playback or record)
 * and place the 8K chunks on the buffers free list.
 */
static int
audioi810_alloc_sample_buffer(struct audioi810_softstate *softstate,
			      sample_buffer_t *buf)
{
    int offs;
    struct bdl_chunk *chunk;

    for (offs = 0; offs < SAMPLEBUF_SIZE; offs += SAMPLEBUF_INTR_CHUNK) {

	chunk = kmem_zalloc(sizeof(struct bdl_chunk), KM_NOSLEEP);
	if (chunk == NULL)
	    return AUDIOI810_FAILURE;

	chunk->next = buf->free;
	buf->free = chunk;

	chunk->dma_bound = FALSE;

	if (ddi_dma_alloc_handle(softstate->dip,
				 &dma_attr,
				 DDI_DMA_DONTWAIT, NULL,
				 &chunk->dma_handle) != DDI_SUCCESS) {
	    chunk->dma_handle = 0;
	    return AUDIOI810_FAILURE;
	}

	if (ddi_dma_mem_alloc(chunk->dma_handle,
			      SAMPLEBUF_INTR_CHUNK,
			      &dev_attr,
			      DDI_DMA_STREAMING,
			      DDI_DMA_DONTWAIT, NULL,
			      &chunk->data, &chunk->reallen,
			      &chunk->acc_handle) != DDI_SUCCESS) {
	    DPRINTF1(AUDI810M_ATTACH, "alloc dma sample buffer failed");
	    return AUDIOI810_FAILURE;
	}
    }

    return AUDIOI810_SUCCESS;
}


/*
 * Release the DMA memory sample buffers allocated with
 * audioi810_alloc_sample_buffer()
 */
static int
audioi810_free_sample_buffer(struct audioi810_softstate *softstate,
			     sample_buffer_t *buf)
{
    struct bdl_chunk *chunk, *next;

    for (chunk = buf->free; chunk != NULL; chunk = next) {
	next = chunk->next;

	if (chunk->acc_handle) {
	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }

	    ddi_dma_mem_free(&chunk->acc_handle);
	}

	if (chunk->dma_handle) {
	    ddi_dma_free_handle(&chunk->dma_handle);
	}
	
	kmem_free(chunk, sizeof(struct bdl_chunk));
    }

    return AUDIOI810_SUCCESS;
}


/*
 * audioi810_pci_find_cap
 *
 * find a given pci capability in the capabilities list data structure
 * in pci configuration space.  If found, returns an offset of that
 * capabilities' configuration registers, or 0 if the capability is
 * not present.
 */
static int
audioi810_pci_find_cap(struct audioi810_softstate *softstate, int id)
{
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint16_t pci_status;
    uint8_t cap_id;
    uint8_t cap_ptr;

    /* do we have a capabilities list data structure? */
    pci_status = pci_config_get16(conf_handle, PCI_CONF_STAT);
    if (!(pci_status & PCI_STAT_CAP))
	return 0;

    cap_ptr = pci_config_get8(conf_handle, PCI_CONF_CAP_PTR);

    /*
     * PCI Bus Power Management Interface Specification, Rev 1.1:
     *
     * 3.1: Values of 00h-3Fh are not valid values for the Cap_Ptr.
     * Cap_Ptr needs to be DWORD aligned.
     */
    if (cap_ptr <= 0x3F || (cap_ptr & 3))
	return 0;

    while (cap_ptr != PCI_CAP_NEXT_PTR_NULL) {
	cap_id = pci_config_get8(conf_handle, cap_ptr);
	if (cap_id == id)
	    break;
	cap_ptr = pci_config_get8(conf_handle, cap_ptr+1);
    }

    return cap_ptr;
}


/*
 * audioi810_pci_device_powerlevel
 * 
 * Returns the device's current pci power level, PCI_PMCSR_D0 ...
 * PCI_PMCSR_D3HOT.
 */
static int
audioi810_pci_device_powerlevel(struct audioi810_softstate *softstate)
{
    uint8_t cap_ptr = audioi810_pci_find_cap(softstate, PCI_CAP_ID_PM);

    /*
     * If the device does not support the PCI power management (PM)
     * capability, the device will always be in D0 state.
     */
    if (!cap_ptr)
	return PCI_PMCSR_D0;

    return pci_config_get16(softstate->pci_conf_handle, cap_ptr + PCI_PMCSR)
	& PCI_PMCSR_STATE_MASK;
}

/*
 * Configure the PCI device, so that the device is ready to 
 * accept I/O requests.  This includes enabling PCI busmaster,
 * and anabling access to the device's I/O registers.
 */
static int
audioi810_pci_setup(struct audioi810_softstate *softstate)
{
    dev_info_t *dip = softstate->dip;
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint16_t pci_status;
    uint16_t pci_vendorid;
    uint16_t pci_deviceid;
    uint8_t  pci_revision;
    uint8_t  pci_intrinfo;
    uint16_t pci_commandreg;
    uint16_t enable_access;
    uint8_t  pm_level;
    char *revstr = "";

    pci_status   = pci_config_get16(conf_handle, PCI_CONF_STAT);
    pci_vendorid = pci_config_get16(conf_handle, PCI_CONF_VENID);
    pci_deviceid = pci_config_get16(conf_handle, PCI_CONF_DEVID);
    pci_revision = pci_config_get8 (conf_handle, PCI_CONF_REVID);
    pci_intrinfo = pci_config_get8 (conf_handle, PCI_CONF_ILINE);    
    
    /* try to find out a nice hardware name for the given pci hw */
    softstate->ich4 = softstate->sis = 0;
    switch (pci_vendorid) {
    case 0x8086: /* intel: */
	switch (pci_deviceid) {
	case 0x2415: revstr = "ICH:82801AA"; break;
	case 0x2425: revstr = "ICH0:82801AB"; break;	/*i820?*/
	case 0x2435: revstr = "ICH2:82801AB"; break;	/*ich3*/
	case 0x2445: revstr = "ICH3:82801BA/M"; break;	/*i815e*/
	case 0x2485: revstr = "ICH3:82801CA/M"; break;	/*i830*/
	case 0x24c5: revstr = "ICH4:82801DB/M";		/*i845*/
	    	     softstate->ich4 = 1;		/*i855*/
		     break;
	case 0x24d5: revstr = "ICH5:82801EB";
	    	     softstate->ich4 = 1;
		     break;
	case 0x7195: revstr = "82443MX"; break;		/*440mx*/
	}
	break;
    case 0x10de: /* nvidia: */
	switch (pci_deviceid) {
	case 0x01b1: revstr = "nForce MCP"; break;
	case 0x006a: revstr = "nForce MCP2"; break;
	case 0x00da: revstr = "nForce MCP3"; break;
	}
	break;
    case 0x1039: /* SiS */
	switch (pci_deviceid) {
	case 0x7012: revstr = "SiS7012";
		     softstate->sis = 1;
		     break;
	}
	break;
    case 0x1022: /* AMD */
	switch (pci_deviceid) {
	case 0x7445: revstr = "AMD768"; break;
	case 0x746d: revstr = "AMD8111"; break;
	}
	break;
    }
    softstate->rev = pci_revision;
    softstate->revstr = revstr;
    
    /*
     * SiS 7012 is slightly incompatible with Intel ICH, it has two
     * registers swapped
     */
    if (softstate->sis) {
	softstate->bm_sr_reg = SIS_BM_SR;
	softstate->bm_picb_reg = SIS_BM_PICB;
	softstate->sample_size = 1;
    } else {
	softstate->bm_sr_reg = I810_BM_SR;
	softstate->bm_picb_reg = I810_BM_PICB;
	softstate->sample_size = 2;
    }

    /* Announce the hw we're attaching to */
    ddi_report_dev(dip);
    cmn_err(CE_CONT, "?PCI-device: %s, audioi810%d, %s%s(rev. 0x%02x) at irq %d\n",
	    ddi_node_name(dip),
	    ddi_get_instance(dip), 
	    revstr, revstr[0] ? " " : "", 
	    pci_revision,
	    pci_intrinfo);

#if i386
    /*
     * reject interrupt line 255, per PCI spec 6.2.4, footnote 43,
     * this means "unknown" or "not connected" to the interrupt
     * controller on x86 based PCs.
     * Sun drivers (iprb) reject iline 0 as well.
     */
    if (pci_intrinfo == 0 || pci_intrinfo == 255) {
	cmn_err(CE_WARN, "interrupt line value out of range: %d", pci_intrinfo);
	return AUDIOI810_FAILURE;
    }
#endif

    pci_commandreg = pci_config_get16(conf_handle, PCI_CONF_COMM);
    DPRINTF4(AUDI810M_ATTACH, "pci command register 0x%x %s%s%s%s",
	    pci_commandreg,
	    pci_commandreg & PCI_COMM_SERR_ENABLE ? " serr#_en"  : "",
	    pci_commandreg & PCI_COMM_ME  ? " busmaster" : "",
	    pci_commandreg & PCI_COMM_MAE ? " mem_access" : "",
	    pci_commandreg & PCI_COMM_IO  ? " io_access" : "");

    enable_access = PCI_COMM_IO|PCI_COMM_ME;
    if (softstate->ich4)
	enable_access |= PCI_COMM_MAE;	/* ICH4 has mmap'ed I/O regs */

    if ((pci_commandreg & enable_access) == enable_access) {
	DPRINTF4(AUDI810M_ATTACH, "busmastering and io_access already enabled");
    } else {
	DPRINTF4(AUDI810M_ATTACH, "Enabling PCI I/O and bus mastering access");
	pci_config_put16(conf_handle, PCI_CONF_COMM, pci_commandreg | enable_access);
    }

    /*
     * If the device isn't in D0 full power state, make sure that
     * there are valid saved pci config registers that we can restore
     * in the power entry point, when the device is powered up.
     */
    pm_level = audioi810_pci_device_powerlevel(softstate);
    if (pm_level != PCI_PMCSR_D0) {
	/*
	cmn_err(CE_WARN, "Device is in PCI D%d power state, "
		"and not in PCI D0 (full power) state.", pm_level);
	*/
	if (pci_save_config_regs(dip) != DDI_SUCCESS) {
	    cmn_err(CE_WARN, "failed to save pci config regs");
	}
	softstate->saved_iline = 
	    pci_config_get8(conf_handle, PCI_CONF_ILINE);
    }

    return AUDIOI810_SUCCESS;
}


/*
 * Serialize access to the AC97 audio mixer registers.
 * Returns AUDIOI810_SUCCESS, when it's OK to read or write one of the
 * AC97 mixer registers.
 */
static int
audioi810_codec_semaphore(struct audioi810_softstate *softstate)
{
    uint32_t any_codec_rdy;
    uint32_t glob_stat;
    uint8_t sema;
    int loop;

    /*
     * before waiting for the codec's semaphore, make sure the codec we're
     * trying to access exists at all.
     */
    any_codec_rdy = I810_GLBSTAT_CODECRDY | I810_GLBSTAT_CODEC2RDY;
    if (softstate->ich4)
	any_codec_rdy |= I810_GLBSTAT_CODEC3RDY;

    glob_stat = I810_BM_GET32(I810_GLOBAL_STATUS);
    if ((glob_stat & any_codec_rdy) == 0) {
	DPRINTF2(AUDI810M_CODEC, "no codecs ready [0x%x]", glob_stat);
	return AUDIOI810_FAILURE;
    }

    for (loop = 0; loop < 100; loop++) {
	sema = I810_BM_GET8(I810_CODEC_SEMA);
	if ((sema & 1) == 0)		/* no codec access in progress? */
	    return AUDIOI810_SUCCESS;
	drv_usecwait(10);
	
    }
    /* Hmm, codec not ready after 1ms? Should not happen */
    DPRINTF2(AUDI810M_CODEC, "semaphore not ready [0x%x][0x%x]",
	     glob_stat, sema);
    return AUDIOI810_SUCCESS;
}


/*
 * Read an AC97 primary codec mixer register
 *
 * Returns TRUE on success with the data read returned via the
 * *data pointer.  Returns FALSE in case the primary codec is not
 * ready any more for some reason.
 */
static int
audioi810_ac97_read(void *closure, uint16_t sub_addr, uint16_t *data)
{
    struct audioi810_softstate *softstate = closure;

    if (audioi810_codec_semaphore(softstate) != AUDIOI810_SUCCESS) {
	*data = 0xffff;
	return FALSE;
    }

    sub_addr += softstate->codec * I810_CODEC_OFFSET;
    *data = I810_AM_GET16(sub_addr);
    return TRUE;
}


/*
 * Write to an AC97 primary codec mixer register.
 *
 * Returns TRUE on success. Returns FALSE in case the primary codec is
 * not ready any more for some reason.
 */
static int
audioi810_ac97_write(void *closure, uint16_t sub_addr, uint16_t value)
{
    struct audioi810_softstate *softstate = closure;

    if (audioi810_codec_semaphore(softstate) != AUDIOI810_SUCCESS)
	return FALSE;

    sub_addr += softstate->codec * I810_CODEC_OFFSET;
    I810_AM_PUT16(sub_addr, value);
    return TRUE;
}


/*
 * Completely disable all ICH audio controller DMA
 */
static void
audioi810_stop_dma(struct audioi810_softstate *softstate)
{
    /* pause bus master (needed for the following reset register) */
    I810_BM_PUT8 (I810_BM_PI_CR,       0x0);
    I810_BM_PUT8 (I810_BM_PO_CR,       0x0);
    I810_BM_PUT8 (I810_BM_MC_CR,       0x0);

    /* and reset the bm registers for a three dma engines */
    I810_BM_PUT8 (I810_BM_PI_CR,       I810_BM_CONTROL_RESETREGS);
    I810_BM_PUT8 (I810_BM_PO_CR,       I810_BM_CONTROL_RESETREGS);
    I810_BM_PUT8 (I810_BM_MC_CR,       I810_BM_CONTROL_RESETREGS);
}


/*
 * reset the ICH audiocontroller's AC'97 link and reset the AC'97
 * codec to default values
 */
static int
audioi810_reset_card(struct audioi810_softstate *softstate,
		     audio_info_t		*audio_info,
		     int			 verbose)
{
    uint32_t glob_ctrl;
    uint32_t glob_status = 0;
    uint32_t any_codec_rdy;
    int loop;
    int codecs;

    /*
     * ICH AC97 link_reset code from 2.6.3.5 of the ICH's programmer's
     * reference manual.
     *
     * stop the AC97 cold reset condition (I810_GLBCTRL_COLDRESET == 0);
     * if the codec is not in a cold reset state, perform a warm reset.
     * Configure PCM Output for 2-channel mode
     */
    glob_ctrl = I810_BM_GET32(I810_GLOBAL_CONTROL);
    glob_ctrl &= ~(I810_GLBCTRL_ACLINKOFF | I810_GLBCTRL_CHANNEL_MASK);
    glob_ctrl |= (glob_ctrl & I810_GLBCTRL_COLDRESET) == 0
	? I810_GLBCTRL_COLDRESET
	: I810_GLBCTRL_WARMRESET;
    I810_BM_PUT32(I810_GLOBAL_CONTROL, glob_ctrl);

    /* wait until the warm reset finishes (not more than 600ms) */
    for (loop = 6000; --loop >= 0; ) {
	drv_usecwait(100);
	glob_ctrl = I810_BM_GET32(I810_GLOBAL_CONTROL);
	if ((glob_ctrl & I810_GLBCTRL_WARMRESET) == 0)
	    break;
    }
    if (loop < 0 ) {
	DPRINTF2(AUDI810M_ATTACH,
		 "AC97 warm reset still in progress? [0x%x]\n", glob_ctrl);
	return AUDIOI810_FAILURE;
    }

    /* 
     * wait for any codec to enter 'codec ready' state (max 750ms).
     *
     * in chapter 5.3 (ICH: Programmer's Reference Manual), Intel
     * recommends a 600ms wait for a codec ready status.
     *
     * Note that the primary (and possible only) ac97 codec may be
     * connected to any of the AC_SDIN[0] .. AC_SDIN[2] lines.
     */
    any_codec_rdy = I810_GLBSTAT_CODECRDY | I810_GLBSTAT_CODEC2RDY;
    if (softstate->ich4)
	any_codec_rdy |= I810_GLBSTAT_CODEC3RDY;

    for (loop = 7500; --loop >= 0; ) {
	drv_usecwait(100);
	glob_status = I810_BM_GET32(I810_GLOBAL_STATUS);
	if (glob_status & any_codec_rdy)
	    break;
    }
    if (loop < 0) {
	DPRINTF2(AUDI810M_ATTACH,
		 "no ac97 codec is ready [0x%x]\n", glob_status);
	return AUDIOI810_FAILURE;
    }

    /* some codec is ready now */

    if ((glob_status & any_codec_rdy) != any_codec_rdy) {
	/*
	 * some codec is ready now, maybe there are more but they
	 * are slow.
	 *
	 * wait for more codecs to enter 'codec ready' state
	 * (including the above wait for the first few codecs, but
	 * not longer than 750ms)
	 */
	while (--loop >= 0) {
	    drv_usecwait(100);
	    glob_status = I810_BM_GET32(I810_GLOBAL_STATUS);
	    if ((glob_status & any_codec_rdy) == any_codec_rdy)
		break;
	}
	if (loop < 0) {
	    DPRINTF3(AUDI810M_ATTACH,
		     "less then %s codecs ready [0x%x]\n",
			softstate->ich4 ? "three" : "two",
			glob_status);
	    /* it's ok when we have less then two (ICH4: three) codecs */
	}
    }

    /*
     * this driver doesn't support fancy multicodec configurations, we're
     * always taking to the primary codec at codec address 00b
     */
    softstate->codec = 0;

    codecs = 0;
    if (glob_status & I810_GLBSTAT_CODEC3RDY) {
	codecs++;
    }
    if (glob_status & I810_GLBSTAT_CODEC2RDY) {
	codecs++;
    }
    if (glob_status & I810_GLBSTAT_CODECRDY) {
	codecs++;
    }

    /*
     * unmute the output on SIS7012
     *
     * XXX: Note: not all SiS7012 mainboards have the 128 bytes busmaster
     * register block (sis == 2), some announce a smaller 64 byte block
     * (sis == 1).
     * Since there's no public specification for the SiS7012, the meaning
     * of the device register at offset 0x4c is unknown. Setting bit 1 of
     * the register is supposed to fix a problem with muted audio.
     */
    if (softstate->sis == 2) {
	I810_BM_PUT16(0x4c, I810_BM_GET16(0x4c) | 1);
    }

    /* no DMA (not yet) */
    audioi810_stop_dma(softstate);

    if (!CODEC_RESET(softstate->codec_handle, audio_info, verbose)) {
	DPRINTF2(AUDI810M_CODEC, "Audio CODEC reset failed");
	return AUDIOI810_FAILURE;
    }

    audioi810_ac97_read (softstate, AC97_EXTENDED_AUDIO_REGISTER, &softstate->ac97_ext_features);
#if 0
#warning software SRC forced
    /* force the use of the software sample rate converter */
    softstate->ac97_ext_features = 0;
#endif

    return AUDIOI810_SUCCESS;
}


/*
 * Allocate DMA resources for the ICH's buffer descriptor lists.  We
 * allocate adjacent DMA memory in one physical page for all three DMA
 * engines available with the ICH audio controller.
 */
static int
audioi810_setup_dma(struct audioi810_softstate *softstate)
{
    int ret;
    uint_t ccount;
    ddi_dma_cookie_t cookie;

    softstate->bdl_dma_bound = FALSE;

    if ((ret = ddi_dma_alloc_handle(softstate->dip, &bdl_dma_attr,
				    DDI_DMA_DONTWAIT, NULL,
				    &softstate->bdl_dma_handle)) != DDI_SUCCESS) {
	DPRINTF4(AUDI810M_ATTACH, "setup_dma: alloc handle failed (%d)\n", ret);
	softstate->bdl_dma_handle = 0;
	return AUDIOI810_FAILURE;
    }

    if (ddi_dma_mem_alloc(softstate->bdl_dma_handle,
			  sizeof(i810_bdl_t),
			  &dev_attr,
			  DDI_DMA_CONSISTENT,
			  DDI_DMA_DONTWAIT, NULL,
			  (caddr_t*) &softstate->bdl,
			  &softstate->bdllen,
			  &softstate->bdl_acc_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDI810M_ATTACH, "setup_dma: mem alloc failed\n");
	return AUDIOI810_FAILURE;
    }

    if ((ret = ddi_dma_addr_bind_handle(softstate->bdl_dma_handle,
					NULL,
					(caddr_t)softstate->bdl, softstate->bdllen,
					DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
					DDI_DMA_DONTWAIT, NULL,
					&cookie, &ccount)) != DDI_DMA_MAPPED) {
	DPRINTF4(AUDI810M_ATTACH, "setup_dma: bind failed (%d)\n", ret);
	return AUDIOI810_FAILURE;
    }

    /*
     * We should receive exactly one physical page mapping
     * (=contiguous physical memory)
     */
    if (ccount != 1) {
	DPRINTF1(AUDI810M_ATTACH, "unexpected dma cookie count > 1 (%u)\n", ccount);
	return AUDIOI810_FAILURE;
    }

    softstate->bdl_dma_bound = TRUE;

    DPRINTF4(AUDI810M_ATTACH, "bdl at physical address %p\n",
	     (void*) (long) cookie.dmac_address);

    softstate->bdl_paddr = (i810_bdl_t*) (long) cookie.dmac_address;

    return AUDIOI810_SUCCESS;
}


/*
 * Cleanup all resources allocated by the driver during attach time
 * and during run-time, in reverse order.
 */
static int
audioi810_cleanup(struct audioi810_softstate *softstate)
{
    if (softstate->audio_handle) {
	if (audio_hlp_detach(softstate->audio_handle, DDI_DETACH) != DDI_SUCCESS)
	    return DDI_FAILURE;
    }

    if (softstate->codec_handle) {
	CODEC_DELETE(softstate->codec_handle);
	softstate->codec_handle = NULL;    
    }

    if (softstate->bdl_acc_handle != NULL) {

	/*
	 * just to be sure, turn off DMA, before releasing the DMA
	 * sample buffers
	 */
	audioi810_stop_dma(softstate);

	/* free audio sample buffer DMA memory */
	audioi810_free_sample_buffer(softstate, &softstate->play_buf);
	audioi810_free_sample_buffer(softstate, &softstate->record_buf);

	/* free BDL DMA memory for the three ICH DMA engines */
	if (softstate->bdl_dma_bound)
	    ddi_dma_unbind_handle(softstate->bdl_dma_handle);
	ddi_dma_mem_free(&softstate->bdl_acc_handle);
	softstate->bdl = NULL;
    }

    if (softstate->bdl_dma_handle) {
	ddi_dma_free_handle(&softstate->bdl_dma_handle);
    }

    /*
     * remove interrupt handler before we free the card's mapped registers
     * otherwise an interrupt could access softstate->regs_handle in our
     * interrupt handler
     */
    if (softstate->intr_cookie != NULL) {
	ddi_remove_intr(softstate->dip, 0, softstate->intr_cookie);
    }

    if (pm_lower_power(softstate->dip, 0, 0) != DDI_SUCCESS) {
	DPRINTF2(AUDI810M_ATTACH, "Failed to disable device power.");
    }

    if (softstate->bm_regs_handle != NULL) {
	ddi_regs_map_free(&softstate->bm_regs_handle);
    }
    if (softstate->am_regs_handle != NULL) {
	ddi_regs_map_free(&softstate->am_regs_handle);
    }

    if (softstate->pci_conf_handle != NULL) {
	pci_config_teardown(&softstate->pci_conf_handle);
    }

    mutex_destroy(&softstate->instance_mutex);
    mutex_destroy(&softstate->intr_mutex);

    ddi_soft_state_free(audioi810_state, ddi_get_instance(softstate->dip));

    return DDI_SUCCESS;
}


/*
 * Playback of a chunk of sample data has finished, release the DMA buffer
 * to the free list.
 */
static void
audioi810_release_play_buf_head(struct audioi810_softstate *softstate,
				sample_buffer_t *buf)
{
    struct bdl_chunk *chunk;

    if ((chunk = buf->active[buf->head]) != NULL) {
	DPRINTF4(AUDI810M_INTR, "free chunk at index %d\n", buf->head);
	if (chunk->dma_bound) {
	    ddi_dma_unbind_handle(chunk->dma_handle);
	    chunk->dma_bound = FALSE;
	}
	chunk->next = buf->free;
	buf->free = chunk;
	buf->active[buf->head] = NULL;
    }
    ++buf->buffers_available;
    if (++buf->head >= I810_NUM_BD)
	buf->head = 0;
}


/*
 * Recoding to a chunk of sample data at the front of the DMA buffer
 * descriptor list has finished.  Send the recorded audio samples to
 * the audio driver framework and release the buffer to the free list.
 */
static void
audioi810_release_record_buf_head(struct audioi810_softstate *softstate,
				  sample_buffer_t *buf)
{
    struct bdl_chunk *chunk;

    if ((chunk = buf->active[buf->head]) != NULL) {
	DPRINTF4(AUDI810M_INTR, "free chunk at index %d\n", buf->head);
	if (chunk->dma_bound) {
	    ddi_dma_unbind_handle(chunk->dma_handle);
	    chunk->dma_bound = FALSE;
	}

	audio_hlp_send_audio(softstate->audio_handle, chunk->data, 0,
			     SAMPLEBUF_INTR_CHUNK >> 2);

	chunk->next = buf->free;
	buf->free = chunk;
	buf->active[buf->head] = NULL;
    }
    ++buf->buffers_available;
    if (++buf->head >= I810_NUM_BD)
	buf->head = 0;
}


/*
 * interrupt handler for the audioi810 device.
 */
static uint_t
audioi810_intr(caddr_t intr_arg)
{
    struct audioi810_softstate *softstate = (void*)intr_arg;
    sample_buffer_t *buf;
    uint32_t glob_status;
    uint16_t sr;
    uint8_t civ;
    int invalid;

    DPRINTF4(AUDI810M_INTR, "audioi810_intr\n");

    if(softstate->power_down) {
	return DDI_INTR_UNCLAIMED;
    }

    glob_status = I810_BM_GET32(I810_GLOBAL_STATUS);
    if ((glob_status & (I810_GLBSTAT_PINIRQ|I810_GLBSTAT_POUTIRQ|I810_GLBSTAT_MICIRQ)) == 0) {
	/* 
	 * shared PCI interrupt?  The ICH audio controller didn't send
	 * an interrupt 
	 */
	return DDI_INTR_UNCLAIMED;
    }

    invalid = 0;

    if (glob_status & I810_GLBSTAT_PINIRQ) {
	/* a PCM in interrupt from the ICH */
	sr = I810_BM_GET16(I810_BM_PI_OFFSET + softstate->bm_sr_reg);
	civ = I810_BM_GET8(I810_BM_PI_CIV);

	DPRINTF4(AUDI810M_INTR, "PCM in irq, SR=0x%x, CIV=%d\n", sr, civ);

	if (civ < I810_NUM_BD) {
	    buf = &softstate->record_buf;

	    /*
	     * send completed record buffers from the head of the
	     * descriptor list to the audio framework and release the
	     * buffers
	     */
	    mutex_enter(&softstate->intr_mutex);
	    while (buf->head != civ)
		audioi810_release_record_buf_head(softstate, buf);
	    mutex_exit(&softstate->intr_mutex);

	    /*
	     * prepare new record buffers and append them at the tail
	     * of the list
	     */
	    audioi810_prepare_record_buffers(softstate);

	    /* clear PCM in interrupt state */
	    I810_BM_PUT16(I810_BM_PI_OFFSET + softstate->bm_sr_reg,
		     ( I810_BM_STATUS_FIFOERR |
		       I810_BM_STATUS_BUFCOMPLETE |
		       I810_BM_STATUS_LASTBUFCOMPLETE ));
	} else
	    invalid++;
    }

    if (glob_status & I810_GLBSTAT_POUTIRQ) {
	/* a PCM out interrupt from the ICH */
	sr = I810_BM_GET16(I810_BM_PO_OFFSET + softstate->bm_sr_reg);
	civ = I810_BM_GET8(I810_BM_PO_CIV);

	DPRINTF4(AUDI810M_INTR, "PCM out irq, SR=0x%x CIV=%d\n", sr, civ);

	if (civ < I810_NUM_BD) {
	    I810_BM_PUT16(I810_BM_PO_OFFSET + softstate->bm_sr_reg,
		     ( I810_BM_STATUS_FIFOERR |
		       I810_BM_STATUS_BUFCOMPLETE |
		       I810_BM_STATUS_LASTBUFCOMPLETE ));
	    buf = &softstate->play_buf;

	    /* free buffers at the head of the buffer desc list */
	    mutex_enter(&softstate->intr_mutex);
	    while (buf->head != civ)
		audioi810_release_play_buf_head(softstate, buf);
	    mutex_exit(&softstate->intr_mutex);

	    if (sr & I810_BM_STATUS_FIFOERR) {
		DPRINTF4(AUDI810M_INTR, "PCM out FIFOERR, signal play shutdown");
		audio_hlp_play_shutdown(softstate->audio_handle, 0);

		mutex_enter(&softstate->intr_mutex);
		while (buf->head != buf->tail)
		    audioi810_release_play_buf_head(softstate, buf);
		mutex_exit(&softstate->intr_mutex);

		if (softstate->play_timeout_id) {
		    untimeout(softstate->play_timeout_id);
		    softstate->play_timeout_id = 0;
		}
	    } else {
		/* prepare new playback buffers and append them at the tail of the list */
		DPRINTF4(AUDI810M_INTR, "fill more play buffers\n");
		if (audioi810_fill_play_buffers(softstate) == 0 &&
		    (sr & I810_BM_STATUS_ONLASTBUF)) {
		    DPRINTF4(AUDI810M_INTR, "out of data, shutdown playback\n");
		    audio_hlp_play_shutdown(softstate->audio_handle, 0);

		    mutex_enter(&softstate->intr_mutex);
		    audioi810_release_play_buf_head(softstate, buf);
		    mutex_exit(&softstate->intr_mutex);
		}
	    }
	} else
	    invalid++;
    }

    if (glob_status & I810_GLBSTAT_MICIRQ) {
	/* a MIC in interrupt (not used in this driver) */
	sr = I810_BM_GET16(I810_BM_MC_OFFSET + softstate->bm_sr_reg);
	civ = I810_BM_GET8(I810_BM_MC_CIV);

	DPRINTF4(AUDI810M_INTR, "MIC in irq, SR=0x%x CIV=%d\n", sr, civ);

	if (civ < I810_NUM_BD) {
	    I810_BM_PUT16(I810_BM_MC_OFFSET + softstate->bm_sr_reg,
		     ( I810_BM_STATUS_FIFOERR |
		       I810_BM_STATUS_BUFCOMPLETE |
		       I810_BM_STATUS_LASTBUFCOMPLETE ));
	} else
	    invalid++;
    }

    /* PI/PO/MV CIV all invalid? I/O space completely disabled? */
    if (invalid == 3 && glob_status == 0xffffffff) {
	cmn_err(CE_WARN, "i/o registers disabled?");
	return DDI_INTR_UNCLAIMED;
    }

    return DDI_INTR_CLAIMED;
}


static audio_device_t audioi810_dev_info;
static ah_ad_entry_t audioi810_entry;
static ah_ad_ch_cap_t audioi810_ch_cap;
static ah_ad_cap_comb_t audioi810_cap_comb[];
static ah_ad_sample_rates_t audioi810_sample_rates_48k;

static char *pmcomps[] = {
    "NAME=PCI Device Power",
    "0=Off (PCI D3 State)",
    "3=On (PCI D0 State)"
};


static int
audioi810_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audioi810_softstate *softstate;
    int status;
    char *getdev_name;
    int *clocks;
    u_int nclocks;
    int am_reg_idx;
    int bm_reg_idx;
    int bm_regs_size;
    int32_t *reg;
    int reg_prop_len;


    DPRINTF4(AUDI810M_ATTACH, "attach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_ATTACH:
	break;

    case DDI_RESUME:
    case DDI_PM_RESUME:
    default:
	return DDI_FAILURE;
    }
	
    /*
     * this card is a bus master, reject any slave-only slot
     */
    if (ddi_slaveonly(dip) == DDI_SUCCESS) {
        cmn_err(CE_NOTE, "slot does not support PCI bus-master");
        return DDI_FAILURE;
    }

    /*
     * Make sure we're not using a high level interrupt handler.
     */
    if (ddi_intr_hilevel(dip, 0)) {
	cmn_err(CE_NOTE, "Driver does not support high level interrupt.");
	cmn_err(CE_NOTE, "You should probably add 'interrupt-priorities=9;' to audioi810.conf");
	return DDI_FAILURE;
    }

    if (ddi_soft_state_zalloc(audioi810_state, instance) != DDI_SUCCESS)
	return DDI_FAILURE;

    softstate = ddi_get_soft_state(audioi810_state, instance);
    softstate->dip = dip;

    if (ddi_prop_update_string_array(DDI_DEV_T_NONE, softstate->dip,
	"pm-components", &pmcomps[0],
	sizeof(pmcomps) / sizeof(pmcomps[0])) != DDI_PROP_SUCCESS)
    {
	DPRINTF4(AUDI810M_ATTACH, "audioi810%d: create pm-components failed",
	    instance);
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (pci_config_setup(softstate->dip, &softstate->pci_conf_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDI810M_ATTACH, "audioi810%d: pci_config_setup failed", instance);
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audioi810_pci_setup(softstate) != AUDIOI810_SUCCESS) {
	DPRINTF4(AUDI810M_ATTACH, "audioi810%d: audioi810_pci_setup failed", instance);
	audioi810_cleanup(softstate);
	return DDI_FAILURE;	
    }

    bcopy(&audioi810_defaults, &softstate->audio_info, sizeof(audio_info_t));
    bcopy(&audioi810_dev_info, &softstate->dev_info, sizeof(audio_device_t));

    /* allow spoofing of audio_device.name in AUDIO_GETDEV ioctl */
    if (ddi_prop_lookup_string(DDI_DEV_T_ANY, softstate->dip,
			       DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
			       "ioctl_audio_getdev_name",
			       &getdev_name) == DDI_PROP_SUCCESS) {

	strncpy(softstate->dev_info.name, getdev_name,
		sizeof(softstate->dev_info.name));
	ddi_prop_free(getdev_name);
    }

    snprintf(softstate->dev_info.version, sizeof(softstate->dev_info.version),
	"%02X,%s", softstate->rev, softstate->revstr);

    softstate->ad_info.ad_int_vers  = AD_VERS1;
    bcopy(&audioi810_defaults, &softstate->ad_defaults, sizeof(audio_info_t));
    softstate->ad_info.ad_defaults  = &softstate->ad_defaults;
    bcopy(&audioi810_ch_cap, &softstate->ad_info.ad_play,   sizeof(ah_ad_ch_cap_t));
    bcopy(&audioi810_ch_cap, &softstate->ad_info.ad_record, sizeof(ah_ad_ch_cap_t));
    softstate->ad_info.ad_play_comb = audioi810_cap_comb;
    softstate->ad_info.ad_rec_comb  = audioi810_cap_comb;
    softstate->ad_info.ad_entry     = &audioi810_entry;
    softstate->ad_info.ad_dev_info  = &softstate->dev_info;
    softstate->ad_info.ad_translate_flags =
				AD_MISC_8_P_TRANSLATE | AD_MISC_8_R_TRANSLATE;

    softstate->audio_handle = audio_hlp_attach(softstate->dip, cmd, &softstate->ad_info);
    if (!softstate->audio_handle) {
	cmn_err(CE_WARN, "failed to attach audio helper module");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    audio_hlp_set_private(softstate->audio_handle, softstate);

    /*
     * map the "native audio mixer" register block.
     *
     * search the "reg" property for the audio mixer register
     * block. On ICH4, we prefer the new memory mapped register block.
     */
    if (ddi_getlongprop(DDI_DEV_T_ANY, softstate->dip, DDI_PROP_DONTPASS,
	"reg", (caddr_t)&reg, &reg_prop_len) != DDI_PROP_SUCCESS)
    {
	cmn_err(CE_WARN, "failed to get 'reg' property");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    for (am_reg_idx = 0;
	 am_reg_idx < reg_prop_len / sizeof(int32_t);
	 am_reg_idx += 5)
    {
	if (softstate->ich4) {
	    if ((reg[am_reg_idx] & PCI_ADDR_MASK) == PCI_ADDR_MEM32
		&& PCI_REG_REG_G(reg[am_reg_idx]) == 0x18) {
		break;
	    }
	} else {
	    if ((reg[am_reg_idx] & PCI_ADDR_MASK) == PCI_ADDR_IO
		&& PCI_REG_REG_G(reg[am_reg_idx]) == 0x10) {
		break;
	    }
	}
    }

    for (bm_reg_idx = 0;
	 bm_reg_idx < reg_prop_len / sizeof(int32_t);
	 bm_reg_idx += 5)
    {
	if (softstate->ich4) {
	    if ((reg[bm_reg_idx] & PCI_ADDR_MASK) == PCI_ADDR_MEM32
		&& PCI_REG_REG_G(reg[bm_reg_idx]) == 0x1c) {
		break;
	    }
	} else {
	    if ((reg[bm_reg_idx] & PCI_ADDR_MASK) == PCI_ADDR_IO
		&& PCI_REG_REG_G(reg[bm_reg_idx]) == 0x14) {
		break;
	    }
	}
    }

    /*
     * Some SiS7012 mainboards have a 64 byte bm reg block, some have a
     * 128 byte block 
     */
    if (softstate->sis && bm_reg_idx < reg_prop_len) {
	if (reg[bm_reg_idx + 3] || reg[bm_reg_idx + 4] >= SIS_BM_REGS_SIZE)
	    softstate->sis = 2;
    }

    kmem_free(reg, reg_prop_len);

    if (am_reg_idx >= reg_prop_len) {
	cmn_err(CE_WARN, "audio mixer register block not found");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }
    if (bm_reg_idx >= reg_prop_len) {
	cmn_err(CE_WARN, "bus master register block not found");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    am_reg_idx /= 5;
    bm_reg_idx /= 5;

    status = ddi_regs_map_setup(softstate->dip, am_reg_idx,
				(caddr_t *)&softstate->am_regs,	0,
				softstate->ich4 ? I845_AM_REGS_SIZE 
						: I810_AM_REGS_SIZE,
				&dev_attr, &softstate->am_regs_handle);

    if (status != DDI_SUCCESS) {
	if (status == DDI_REGS_ACC_CONFLICT) {
	    cmn_err(CE_WARN, "audio mixer regs mapping conflicted with other mapping");
	} else {
	    cmn_err(CE_WARN, "audio mixer regs mapping failed");
	}
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    /* map the "native audio bus mastering" register block */
    bm_regs_size = softstate->ich4     ? I845_BM_REGS_SIZE
		 : softstate->sis == 2 ? SIS_BM_REGS_SIZE
		 		       : I810_BM_REGS_SIZE;
    status = ddi_regs_map_setup(softstate->dip, bm_reg_idx,
				(caddr_t *)&softstate->bm_regs,
				0, bm_regs_size,
				&dev_attr, &softstate->bm_regs_handle);
    if (status != DDI_SUCCESS) {
	if (status == DDI_REGS_ACC_CONFLICT) {
	    cmn_err(CE_WARN, "bus master regs mapping conflicted with other mapping");
	} else {
	    cmn_err(CE_WARN, "bus master regs mapping failed");
	}
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * enable device power, get the device out of D3hot into D0 PCI power
     * management state.
     */
    if (pm_raise_power(softstate->dip, 0, 3) != DDI_SUCCESS) {
	cmn_err(CE_WARN, "Failed to enable device power.");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * Verify that we now have working device registers.
     *
     * On some systems (Toshiba Tecra S1), warm-rebooting from Windows
     * to Solaris results in a dead register block, which confuses us.
     * Fail with an error message now, instead of locking up in the
     * interrupt handler.
     */
    if (I810_BM_GET8(I810_BM_PI_CIV) >= I810_NUM_BD ||
	I810_BM_GET8(I810_BM_PO_CIV) >= I810_NUM_BD ||
	I810_BM_GET8(I810_BM_MC_CIV) >= I810_NUM_BD)
    {
	cmn_err(CE_WARN, "bus master regs contain unexpected data (%x,%x,%x)",
		I810_BM_GET8(I810_BM_PI_CIV),
		I810_BM_GET8(I810_BM_PO_CIV),
		I810_BM_GET8(I810_BM_MC_CIV));
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);

    if (ddi_get_iblock_cookie(softstate->dip, 0, &softstate->intr_cookie) != DDI_SUCCESS) {
	cmn_err(CE_WARN, "Could not get iblock cookie");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }
    mutex_init(&softstate->intr_mutex, NULL,
	       MUTEX_DRIVER, (void*)softstate->intr_cookie);

    if (ddi_add_intr(softstate->dip, 0,
		     &softstate->intr_cookie, NULL,
		     audioi810_intr, (caddr_t)softstate) != DDI_SUCCESS) {

	cmn_err(CE_WARN, "Could not add interrupt handler");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * try to reset the audio controller, this verifies basic operation of the
     * sound hardware.
     */

    {
	static struct ac97_codec_ops codec_ops = {
	    audioi810_ac97_read,
	    audioi810_ac97_write
	};

	softstate->codec_handle = ac97_codec_init(softstate->audio_handle, &codec_ops, softstate);
    }

    if (audioi810_reset_card(softstate, &audioi810_defaults, TRUE) != AUDIOI810_SUCCESS) {
	cmn_err(CE_WARN, "reset of i810 audio controller failed");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;	
    }


    if (audioi810_setup_dma(softstate) != AUDIOI810_SUCCESS) {
	cmn_err(CE_WARN, "dma setup of i810 audio controller failed");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;	
    }

    if (audioi810_alloc_sample_buffer(softstate, &softstate->play_buf) != AUDIOI810_SUCCESS
	|| audioi810_alloc_sample_buffer(softstate, &softstate->record_buf) != AUDIOI810_SUCCESS) {
	cmn_err(CE_WARN, "cannot allocate sample buffers");
	audioi810_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * It appears there's HW out there, where the AC'97 codec does not
     * run at the 48kHz rate defined by the AC'97 standard.
     *
     * Some HW manufactures save a few cents by not including a
     * separate 24.576 MHz oscillator for the AC'97 codec clock input,
     * from which the internal 48kHz rate is derived.
     * 
     * The integer property "ac97_codec_clockrate" can be used by the
     * user to specify a non-standard internal AC'97 codec sample rate.
     */
    softstate->ac97_clock = 48000;	/* std value, unless overwritten */
    if (ddi_prop_lookup_int_array(DDI_DEV_T_ANY, softstate->dip,
				  DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
				  "ac97_codec_clockrate",
				  &clocks, &nclocks) == DDI_PROP_SUCCESS) {
	if (instance < nclocks
	    && clocks[instance] > 0 && clocks[instance] < 65536)
	    softstate->ac97_clock = clocks[instance];
	ddi_prop_free(clocks);
	cmn_err(CE_NOTE, "AC'97 internal clock rate: %d Hz",
		softstate->ac97_clock);
    }

    /*
     * if we have an AC'97 codec without sample rate converter, 48kHz
     * is all we can play/record
     */
    if (!(softstate->ac97_ext_features & EAR_VRA)) {
	cmn_err(CE_NOTE, "no hardware sample rate converter available");
	bcopy(&audioi810_sample_rates_48k, &softstate->ad_info.ad_play.ad_srs,
	      sizeof(ah_ad_sample_rates_t));
	bcopy(&audioi810_sample_rates_48k, &softstate->ad_info.ad_record.ad_srs,
	      sizeof(ah_ad_sample_rates_t));
    }

    return DDI_SUCCESS;
}


static int
audioi810_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audioi810_softstate *softstate = ddi_get_soft_state(audioi810_state, instance);

    DPRINTF4(AUDI810M_ATTACH, "detach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_DETACH:
	break;

    default:
	return DDI_FAILURE;
    }

    if (audioi810_cleanup(softstate) != DDI_SUCCESS) 
	return DDI_FAILURE;

    return DDI_SUCCESS;
}


static int
audioi810_power(dev_info_t *dip, int component, int level)
{
    int instance = ddi_get_instance(dip);
    struct audioi810_softstate *softstate =
	ddi_get_soft_state(audioi810_state, instance);
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint8_t cap_ptr;

    DPRINTF3(AUDI810M_ATTACH, "power, component=%d, level=%d\n",
	component, level);

    if (component != 0 || (level != 0 && level != 3))
	return DDI_FAILURE;

    if (!conf_handle)
	return DDI_FAILURE;

    cap_ptr = audioi810_pci_find_cap(softstate, PCI_CAP_ID_PM);
    if (!cap_ptr)
	return level == 3 ? DDI_SUCCESS : DDI_FAILURE;

    switch (level) {
    case 0: /* off, PCI D3 */
	if (audioi810_pci_device_powerlevel(softstate) == PCI_PMCSR_D3HOT)
	    break;

	pci_save_config_regs(dip);
	softstate->saved_iline = 
	    pci_config_get8(conf_handle, PCI_CONF_ILINE);
	pci_config_put16(conf_handle, cap_ptr + PCI_PMCSR, PCI_PMCSR_D3HOT);
	softstate->power_down = 1;
	break;

    case 3: /* on, PCI D0 */
	if (audioi810_pci_device_powerlevel(softstate) == PCI_PMCSR_D0)
	    break;

	pci_config_put16(conf_handle, cap_ptr + PCI_PMCSR, PCI_PMCSR_D0);
	/* 10 ms delay for transition from D3hot -> D0 */
	delay(drv_usectohz(10000));
	pci_restore_config_regs(dip);
	pci_config_put8(conf_handle, PCI_CONF_ILINE, softstate->saved_iline);
	softstate->power_down = 0;
	break;
    }
    return DDI_SUCCESS;
}


/*
 * Audio Driver functions (for ah_ad_entry_t)
 */

static int    
audioi810_setup(audiohdl_t handle, int stream, int flags)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDI810M_DRIVER, "audioi810_setup, stream %d, flags %x\n", stream, flags);

    softstate->play_buf.io_started = FALSE;
    softstate->record_buf.io_started = FALSE;

    return AUDIO_SUCCESS;
}


static void
audioi810_teardown(audiohdl_t handle, int stream, int dir)
{
    /* struct audioi810_softstate *softstate = audio_hlp_get_private(handle); */

    DPRINTF4(AUDI810M_DRIVER, "audioi810_teardown, stream %d, dir %d\n", stream, dir);
}


static int
audioi810_set_config(audiohdl_t handle, int stream,
		     int command, int dir, int arg1, int arg2)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    
    DPRINTF4(AUDI810M_DRIVER, "audioi810_set_config, stream %d, "
	    "command %d, dir %d, arg1 %d, arg2 %d\n",
	    stream, command, dir, arg1, arg2);

    pm_raise_power(softstate->dip, 0, 3);
    pm_busy_component(softstate->dip, 0);
    CODEC_SET_CONFIG(softstate->codec_handle, command, dir, arg1, arg2);
    pm_idle_component(softstate->dip, 0);

    return AUDIO_SUCCESS;
}


#ifndef	AC97_EXTENDED_LR_ADC_RATE_REGISTER
/* workaround a bug in sun's ac97.h header file: */
#define	AC97_EXTENDED_LR_ADC_RATE_REGISTER	AC97_EXTENDED_LR_DAC_RATE_REGISTER
#endif


static int
audioi810_set_format(audiohdl_t handle, int stream,
		 int dir, int sample_rate, int channels,
		 int precision, int encoding)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    uint16_t val;
    int rval;

    DPRINTF4(AUDI810M_DRIVER, "audioi810_set_format, stream %d, dir %d, "
	    "sample_rate %d, channels %d, precision %d, encoding %d\n",
	    stream, dir, sample_rate, channels, precision, encoding);

    if (!(softstate->ac97_ext_features & EAR_VRA) && 
	sample_rate != AC97_SAMPLE_RATE_48000) {
	DPRINTF2(AUDI810M_DRIVER, "unsupported sample rate %d\n", sample_rate);
	return AUDIO_FAILURE;
    }

    /* do not fail, otherwise the driver will not unload */
    rval = AUDIO_SUCCESS;

    pm_raise_power(softstate->dip, 0, 3);
    pm_busy_component(softstate->dip, 0);

    /*
     * Workaround for i810 with AC'97 codec running at a non-standard
     * clock speed > 48000 Hz; these devices play sound too fast.
     */
    sample_rate = (unsigned)sample_rate * 48000U / softstate->ac97_clock;

    audioi810_ac97_write(softstate, AC97_EXTENDED_AUDIO_STAT_CTRL_REGISTER, EASCR_VRA);
    if (dir == AUDIO_PLAY) {
	DPRINTF4(AUDI810M_DRIVER, "set DAC VRA %d\n", sample_rate);
	audioi810_ac97_write(softstate, AC97_EXTENDED_FRONT_DAC_RATE_REGISTER, sample_rate);
	audioi810_ac97_write(softstate, AC97_EXTENDED_SURROUND_DAC_RATE_REGISTER, sample_rate);
	audioi810_ac97_write(softstate, AC97_EXTENDED_LFE_DAC_RATE_REGISTER, sample_rate);

	audioi810_ac97_read (softstate, AC97_EXTENDED_FRONT_DAC_RATE_REGISTER, &val);
	DPRINTF4(AUDI810M_DRIVER, "Front DAC samplerate is %d\n", val);
	/*
	 * the whole "extended audio" register block is optional, and may read
	 * as "0".  In this case the AC'97 supports 48kHz only.
	 */
	if (val != 0 && val != sample_rate) {
	    DPRINTF2(AUDI810M_DRIVER,
		     "AC'97 codec does not support playback sample rate %d, uses %d\n",
		     sample_rate, val);
	    rval = AUDIO_FAILURE;
	}
    } else {
	DPRINTF4(AUDI810M_DRIVER, "set ADC VRA %d\n", sample_rate);
	audioi810_ac97_write(softstate, AC97_EXTENDED_LR_ADC_RATE_REGISTER, sample_rate);
	audioi810_ac97_write(softstate, AC97_EXTENDED_MIC_ADC_RATE_REGISTER, sample_rate);
	audioi810_ac97_read (softstate, AC97_EXTENDED_LR_ADC_RATE_REGISTER, &val);
	DPRINTF4(AUDI810M_DRIVER, "LR ADC samplerate is %d\n", val);
	/*
	 * the whole "extended audio" register block is optional, and may read
	 * as "0".  In this case the AC'97 supports 48kHz only.
	 */
	if (val != 0 && val != sample_rate) {
	    DPRINTF2(AUDI810M_DRIVER,
		     "AC'97 codec does not support record sample rate %d, uses %d\n",
		     sample_rate, val);
	    rval = AUDIO_FAILURE;
	}
    }
    pm_idle_component(softstate->dip, 0);

    return rval;
}


static void
audioi810_timed_fill_play_buf(void *arg)
{
    struct audioi810_softstate *softstate = arg;

    DPRINTF4(AUDI810M_DRIVER, "audioi810_timed_fill_play_buf\n");

    mutex_enter(&softstate->intr_mutex);
    softstate->play_timeout_id = 0;
    mutex_exit(&softstate->intr_mutex);

    audioi810_fill_play_buffers(softstate);    
}


/*
 * Fetch free sample buffers from the free list, fill them with audio
 * samples obtained from the audio framework and add them at the tail
 * of the ICH's BDL.
 */
static int
audioi810_fill_play_buffers(struct audioi810_softstate *softstate)
{
    struct bdl_chunk *chunk;
    struct sample_buffer *buf = &softstate->play_buf;
    int bytes, len;
    int total_samples;
    int samples;
    int bindstat;
    int i;
    uint_t ccount;
    ddi_dma_cookie_t cookie;
    i810_buf_descr_t *bdesc;
    uint8_t cr;

    total_samples = 0;

    mutex_enter(&softstate->intr_mutex);

    if (!buf->io_started) {
	I810_BM_PUT8 (I810_BM_PO_CR,       0);
	I810_BM_PUT8 (I810_BM_PO_CR,       I810_BM_CONTROL_IOCENABLE|I810_BM_CONTROL_RESETREGS);
	/* 
	 * XXX i855PM, with memory mapped registers:
	 * delay writing the DESCLIST, while the reset runs? 
	 */
	drv_usecwait(1);

	I810_BM_PUT32(I810_BM_PO_DESCLIST, (uint32_t) (long) softstate->bdl_paddr->pcm_out_bdl);
	I810_BM_PUT8 (I810_BM_PO_LVI,      0);

	buf->head = 0;
	buf->tail = 0;
	buf->buffers_available = I810_NUM_BD;
    }

    while ((chunk = buf->free) != NULL) {
	buf->free = chunk->next;
	chunk->next = NULL;

	bytes = SAMPLEBUF_INTR_CHUNK;
	samples = bytes >> 2;

	samples = audio_hlp_get_audio(softstate->audio_handle, chunk->data, 0, samples);
	DPRINTF4(AUDI810M_INTR, "got %d samples audio data\n", samples);
	if (samples <= 0) {
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}
	bytes = samples << 2;
	total_samples += samples;

	bindstat = ddi_dma_addr_bind_handle(chunk->dma_handle,
					    NULL,
					    chunk->data, chunk->reallen,
					    DDI_DMA_WRITE,
					    DDI_DMA_DONTWAIT, NULL,
					    &cookie, &ccount);
	if (bindstat != DDI_DMA_MAPPED) {
	    DPRINTF4(AUDI810M_INTR, "dma_addr_bind_handle failed %d\n", bindstat);
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}

	chunk->dma_bound = TRUE;
	buf->active[buf->tail] = chunk;

	for (i = 0; i < ccount && bytes > 0; i++) {
	    if (i > 0)
		ddi_dma_nextcookie(chunk->dma_handle, &cookie);
	    DPRINTF4(AUDI810M_INTR, "cookie #%d: at %p, %ld bytes\n",
		     i, (void*) (long)cookie.dmac_address, (long)cookie.dmac_size);

	    if (cookie.dmac_size < bytes)
		len = cookie.dmac_size;
	    else
		len = bytes;
	    bytes -= len;

	    buf->buffers_available--;

	    /*
	     * since we map DMA memory with the DDI_NEVERSWAP_ACC attribute, 
	     * we can access it without using ddi_{get,put}N
	     */
	    bdesc = softstate->bdl->pcm_out_bdl + buf->tail;
	    bdesc->bd_ptr = cookie.dmac_address;
	    bdesc->bd_len = len / softstate->sample_size;
	    bdesc->bd_bup = 0;
	    bdesc->bd_ioc = (bytes <= 0);

	    I810_BM_PUT8(I810_BM_PO_LVI, buf->tail);
	    DPRINTF4(AUDI810M_INTR, "LVI=%d filled\n", buf->tail);
	    if (++buf->tail >= I810_NUM_BD)
		buf->tail = 0;
	}
    }

    cr = I810_BM_GET8(I810_BM_PO_CR);
    if (!buf->io_started) {
	cr &= ~(I810_BM_CONTROL_FEIENABLE|I810_BM_CONTROL_LVBENABLE);
	cr |= I810_BM_CONTROL_IOCENABLE;
	buf->io_started = TRUE;
	pm_busy_component(softstate->dip, 0);
    }
    if (buf->head != buf->tail)
	cr |= I810_BM_CONTROL_DMASTART;
    I810_BM_PUT8(I810_BM_PO_CR, cr);

    if (buf->free && !softstate->play_timeout_id) {
	/*
	 * free sample buffers left, and no more sample data available
	 * from the framework. Schedule a timeout routine to retry
	 * filling more buffers in time before the ICH has used all
	 * the samples we added to the BDL.
	 * 
	 * XXX The timeout (20ms) could be tuned a bit.  At 48kHz,
	 * 16bit, stereo, an 8K sample buffer plays ~41ms. At lower
	 * sample rates the timeout could be reduced.
	 */
	softstate->play_timeout_id = 
	    timeout(audioi810_timed_fill_play_buf, softstate,
		    drv_usectohz(20000));
    }

    mutex_exit(&softstate->intr_mutex);

    DPRINTF4(AUDI810M_INTR,
	     "DMA enabled, PCM out CIV=%d LVI=%d SR=0x%x CR=0x%x\n",
	     I810_BM_GET8(I810_BM_PO_CIV),
	     I810_BM_GET8(I810_BM_PO_LVI),
	     I810_BM_GET16(I810_BM_PO_OFFSET + softstate->bm_sr_reg),
	     I810_BM_GET8(I810_BM_PO_CR));

    return total_samples;
}


/*
 * Fetch free sample buffers from the free list, and add them at the
 * tail of the ICH's BDL for recording audio samples.
 */
static int
audioi810_prepare_record_buffers(struct audioi810_softstate *softstate)
{
    struct bdl_chunk *chunk;
    struct sample_buffer *buf = &softstate->record_buf;
    int bindstat;
    int i;
    uint_t ccount;
    ddi_dma_cookie_t cookie;
    i810_buf_descr_t *bdesc;
    uint8_t cr;

    mutex_enter(&softstate->intr_mutex);

    if (!buf->io_started) {
	I810_BM_PUT8 (I810_BM_PI_CR,       0);
	I810_BM_PUT8 (I810_BM_PI_CR,       I810_BM_CONTROL_IOCENABLE|I810_BM_CONTROL_RESETREGS);
	/* 
	 * XXX i855PM, with memory mapped registers:
	 * delay writing the DESCLIST, while the reset runs? 
	 */
	drv_usecwait(1);

	I810_BM_PUT32(I810_BM_PI_DESCLIST, (uint32_t) (long) softstate->bdl_paddr->pcm_in_bdl);
	I810_BM_PUT8 (I810_BM_PI_LVI,      0);

	buf->head = 0;
	buf->tail = 0;
	buf->buffers_available = I810_NUM_BD;
    }
    
    while ((chunk = buf->free) != NULL) {
	buf->free = chunk->next;
	chunk->next = NULL;

	bindstat = ddi_dma_addr_bind_handle(chunk->dma_handle,
					    NULL,
					    chunk->data, chunk->reallen,
					    DDI_DMA_READ,
					    DDI_DMA_DONTWAIT, NULL,
					    &cookie, &ccount);
	if (bindstat != DDI_DMA_MAPPED) {
	    DPRINTF4(AUDI810M_INTR, "dma_addr_bind_handle failed %d\n", bindstat);
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}

	chunk->dma_bound = TRUE;
	buf->active[buf->tail] = chunk;

	for (i = 0; i < ccount; i++) {
	    if (i > 0)
		ddi_dma_nextcookie(chunk->dma_handle, &cookie);
	    DPRINTF4(AUDI810M_INTR, "cookie #%d: at %p, %ld bytes\n",
		     i, (void*) (long)cookie.dmac_address, (long)cookie.dmac_size);

	    buf->buffers_available--;

	    /*
	     * since we map DMA memory with the DDI_NEVERSWAP_ACC attribute, 
	     * we can access it without using ddi_{get,put}N
	     */
	    bdesc = softstate->bdl->pcm_in_bdl + buf->tail;
	    bdesc->bd_ptr = cookie.dmac_address;
	    bdesc->bd_len = cookie.dmac_size / softstate->sample_size;
	    bdesc->bd_bup = 0;
	    bdesc->bd_ioc = (i == ccount-1);

	    I810_BM_PUT8(I810_BM_PI_LVI, buf->tail);
	    DPRINTF4(AUDI810M_INTR, "LVI=%d filled\n", buf->tail);
	    if (++buf->tail >= I810_NUM_BD)
		buf->tail = 0;
	}
    }

    if (!buf->io_started) {

	cr = I810_BM_GET8(I810_BM_PI_CR);
	cr &= ~(I810_BM_CONTROL_FEIENABLE|I810_BM_CONTROL_LVBENABLE);
	cr |= I810_BM_CONTROL_IOCENABLE|I810_BM_CONTROL_DMASTART;
	I810_BM_PUT8(I810_BM_PI_CR, cr);

	buf->io_started = TRUE;
	pm_busy_component(softstate->dip, 0);
    }

    mutex_exit(&softstate->intr_mutex);

    return AUDIOI810_SUCCESS;
}


/*
 * Audio framework has seen the first audio samples for playback,
 * or playback resumes after a 'pause_play'. 
 */
static int
audioi810_start_play(audiohdl_t handle, int stream)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDI810M_DRIVER, "audioi810_start_play, stream %d\n", stream);
    
    pm_raise_power(softstate->dip, 0, 3);
    audioi810_fill_play_buffers(softstate);

    return AUDIO_SUCCESS;
}


static void
audioi810_pause_play(audiohdl_t handle, int stream)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    uint8_t cr;

    DPRINTF4(AUDI810M_DRIVER, "audioi810_pause_play, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);

    cr = I810_BM_GET8(I810_BM_PO_CR);
    cr &= ~I810_BM_CONTROL_DMASTART;
    I810_BM_PUT8(I810_BM_PO_CR, cr);
    
    mutex_exit(&softstate->intr_mutex);
}


static void
audioi810_stop_play(audiohdl_t handle, int stream)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    int i;
    struct bdl_chunk *chunk;
    sample_buffer_t *buf = &softstate->play_buf;

    DPRINTF4(AUDI810M_DRIVER, "audioi810_stop_play, stream %d\n", stream);

    if (softstate->play_timeout_id) {
	untimeout(softstate->play_timeout_id);
	softstate->play_timeout_id = 0;
    }

    mutex_enter(&softstate->intr_mutex);
    softstate->play_buf.io_started = FALSE;
    pm_idle_component(softstate->dip, 0);

    /* pause bus master */
    I810_BM_PUT8(I810_BM_PO_CR, 0);
    /* reset registers */
    I810_BM_PUT8(I810_BM_PO_CR, I810_BM_CONTROL_RESETREGS);

    /*
     * throw away the DMA chunks that were active on the ICH "PCM out"
     * DMA engine, release them to the free list
     */
    for (i = 0; i < I810_NUM_BD; i++) {
	if ((chunk = buf->active[i]) != NULL) {
	    buf->active[i] = NULL;

	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }
	    chunk->next = buf->free;
	    buf->free = chunk;
	}
    }

    mutex_exit(&softstate->intr_mutex);
}


static int
audioi810_get_delay(audiohdl_t handle, int stream)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    int samples;
    uint8_t idx;

    DPRINTF4(AUDI810M_DRIVER, "audioi810_get_delay, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);
    if (softstate->play_buf.io_started) {
	/* ICH: # Words / SiS: ??? # Bytes? */
	samples = I810_BM_GET16(I810_BM_PO_OFFSET + softstate->bm_picb_reg);
	idx = I810_BM_GET8(I810_BM_PO_CIV);
	/*
	{
	  uint8_t idx2;
	  idx2 = I810_BM_GET8(I810_BM_PO_PIV);
	  cmn_err(CE_NOTE, "^%d/%d %d\n", idx, idx2, samples);
	}
	*/
	for (;;) {
	    idx = (idx+1) % I810_NUM_BD;
	    if (idx == softstate->play_buf.tail)
		break;
	    samples += softstate->bdl->pcm_out_bdl[idx].bd_len; /* # Words */
	}
	/* convert to # samples */
	samples = (samples * softstate->sample_size) >> 2;
    } else
	samples = 0;
    mutex_exit(&softstate->intr_mutex);

    return samples;
}


static int
audioi810_start_record(audiohdl_t handle, int stream)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDI810M_DRIVER, "audioi810_start_record, stream %d\n", stream);

    pm_raise_power(softstate->dip, 0, 3);
    audioi810_prepare_record_buffers(softstate);

    return AUDIO_SUCCESS;
}


static void
audioi810_stop_record(audiohdl_t handle, int stream)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    int i;
    struct bdl_chunk *chunk;
    sample_buffer_t *buf = &softstate->record_buf;
    
    DPRINTF4(AUDI810M_DRIVER, "audioi810_stop_record, stream %d\n", stream);


    mutex_enter(&softstate->intr_mutex);
    softstate->record_buf.io_started = FALSE;
    pm_idle_component(softstate->dip, 0);

    /* pause bus master */
    I810_BM_PUT8(I810_BM_PI_CR, 0);
    /* reset registers */
    I810_BM_PUT8(I810_BM_PI_CR, I810_BM_CONTROL_RESETREGS);

    /*
     * throw away the DMA chunks that were active on the ICH "PCM in"
     * DMA engine, release them to the free list
     */
    for (i = 0; i < I810_NUM_BD; i++) {
	if ((chunk = buf->active[i]) != NULL) {
	    buf->active[i] = NULL;

	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }
	    chunk->next = buf->free;
	    buf->free = chunk;
	}
    }

    mutex_exit(&softstate->intr_mutex);

}


static int
audioi810_ioctl(audiohdl_t handle, int channel,
		queue_t *q, mblk_t *mp, int *error)
{
    struct audioi810_softstate *softstate = audio_hlp_get_private(handle);
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;
    mblk_t *mp2;
    caddr_t arg = 0;
    int vol;
    int gpr_bit;
    uint16_t codecval;

    if (iocp->ioc_count == TRANSPARENT && (mp2 = mp->b_cont) != NULL)
	/* get the "arg" value from the streams msg */
	bcopy(mp2->b_rptr, &arg, sizeof(arg));

    DPRINTF4(AUDI810M_DRIVER, "audioi810_ioctl ioc_cmd=%x, arg=%p\n", iocp->ioc_cmd, arg);

    switch (iocp->ioc_cmd) {
    case AUDIOI810_SET_PCMOUT_VOL:
	vol = ac97_codec_play_vol((int) (long)arg, AUDIO_MAX_GAIN, PCMOVR_RIGHT_GAIN_MASK);
	codecval = (vol << 8) | vol;
	audioi810_ac97_write(softstate, AC97_PCM_OUT_VOLUME_REGISTER, codecval);
	DPRINTF4(AUDI810M_AC97, "PCMVOL=%x", codecval);
	return AD_ACK;

    case AUDIOI810_SET_BEEP_VOL:
	vol = ac97_codec_play_vol((int) (long)arg, AUDIO_MAX_GAIN, 15);
	codecval = vol << 1;
	audioi810_ac97_write(softstate, AC97_PC_BEEP_REGISTER, codecval);
	DPRINTF4(AUDI810M_AC97, "PCBEEPVOL=%x", codecval);
	return AD_ACK;

    case AUDIOI810_SET_LOUDNESS_ENABLE:
	gpr_bit = GPR_BASS_BOOST;
	goto set_gpr_register;

    case AUDIOI810_SET_3D_STEREO_ENABLE:
	gpr_bit = GPR_3D_STEREO_ENHANCE;
	goto set_gpr_register;

    case AUDIOI810_SET_STEREO_ENHANCE_ENABLE:
	gpr_bit = GPR_ST;

    set_gpr_register:	
	audioi810_ac97_read(softstate, AC97_GENERAL_PURPOSE_REGISTER, &codecval);
	if ((long)arg)
	    codecval |= gpr_bit;
	else
	    codecval &= ~gpr_bit;
	audioi810_ac97_write(softstate, AC97_GENERAL_PURPOSE_REGISTER, codecval);
	return AD_ACK;

    case AUDIOI810_SET_3D_STEREO_DEPTH:
	audioi810_ac97_read(softstate, AC97_THREE_D_CONTROL_REGISTER, &codecval);
	codecval &= ~TDCR_DEPTH_MASK;
	codecval |= (long)arg & TDCR_DEPTH_MASK;
	    
	audioi810_ac97_write(softstate, AC97_THREE_D_CONTROL_REGISTER, codecval);
	return AD_ACK;

    case AUDIOI810_SET_3D_STEREO_CENTER:
	audioi810_ac97_read(softstate, AC97_THREE_D_CONTROL_REGISTER, &codecval);
	codecval &= ~TDCR_CENTER_MASK;
	codecval |= (long)arg & TDCR_CENTER_MASK;
	    
	audioi810_ac97_write(softstate, AC97_THREE_D_CONTROL_REGISTER, codecval);
	return AD_ACK;


    default:
	return AD_NACK;
    }
}


static int
audioi810_iocdata(audiohdl_t handle, int channel,
		  queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    DPRINTF4(AUDI810M_DRIVER, "audioi810_iocdata cp_cmd=%x, arg=%p\n", copyresp->cp_cmd, arg);

    switch (copyresp->cp_cmd) {
    default:
	return AD_NACK;
    }
}


static audio_info_t audioi810_defaults = {
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	/*AUDIO_MID_GAIN*/ 255*75/100,		/* gain level: 0 - 255 */
	AUDIO_SPEAKER,		/* selected I/O port */
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_MICROPHONE,		/* selected I/O port */
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    AUDIO_MIN_GAIN,		/* monitor_gain */
    0,				/* output_muted */
#ifdef	AUDIOIO_HAS_REFCNT
    0,				/* driver reference count */
#endif
    {0, 0},			/* reserved fields */
#ifdef	AUDIOIO_HAS_FEATURES
    AUDIO_HWFEATURE_DUPLEX | AUDIO_HWFEATURE_IN2OUT | \
    	AUDIO_HWFEATURE_PLAY | AUDIO_HWFEATURE_RECORD,
    AUDIO_SWFEATURE_MIXER,
    0,
#endif
};


/*
 * Audio Driver ops vector definition
 */
static ah_ad_entry_t audioi810_entry = {
    audioi810_setup,
    audioi810_teardown,
    audioi810_set_config,
    audioi810_set_format,
    audioi810_start_play,
    audioi810_pause_play,
    audioi810_stop_play,
    audioi810_start_record,
    audioi810_stop_record,
    audioi810_ioctl,
    audioi810_iocdata,
    audioi810_get_delay,
};


static audio_device_t audioi810_dev_info = {
    "TOOLS,i810",		/* name */
    "1.0",			/* version */
    "i810"			/* config */
};


/* 0 terminated list of supported sample rates (used as limits) */
static uint_t audioi810_srs[] = {
    8000, 48000, 0
};


/* 0 terminated list of supported sample rates */
static uint_t audioi810_srs_48k[] = {
    AC97_SAMPLE_RATE_48000, 0
};

static ah_ad_sample_rates_t audioi810_sample_rates_48k = {
    AD_SRS_FLAG_SR_NOT_LIMITS, audioi810_srs_48k
};

/*
 * 0 terminated list of supported channels. The ICH AC'97 controller only
 * supports 16-bit signed PCM stereo in the hardware.
 */
static uint_t audioi810_chs[] = {
    AUDIO_CHANNELS_STEREO,
    0
};

static ah_ad_cap_comb_t audioi810_cap_comb[] = {
    { AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR },  /* 16-bit signed PCM */
    { 0 },
};



static ah_ad_ch_cap_t audioi810_ch_cap = {
    { AD_SRS_FLAG_SR_LIMITS, audioi810_srs },
    audioi810_chs,		/* list of channel types */
    0,				/* interrupt rate */
    2,				/* max channels */
    SAMPLEBUF_SIZE,		/* buffer size */
};
