/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if	_KERNEL

#include <sys/sunddi.h>
#include <sys/ksynch.h>
#include <sys/note.h>
#include <sys/mixer.h>
#include <sys/audio.h>
#include <sys/audio/audio_helper.h>


#define	SAMPLEBUF_INTR_CHUNK	0x2000	/* one intr every 8k data */
#define	SAMPLEBUF_SIZE		(3*SAMPLEBUF_INTR_CHUNK)

#define	I810_NUM_BD	32	/* maxium length of buffer descriptor list */


/*
 * An 8K (SAMPLEBUF_INTR_CHUNK) chunk of data of DMA-able memory
 * for which we generate one interrupt from the ICH audio controller.
 */
struct bdl_chunk {
    caddr_t		data;		/* kernel vaddr of data */

    ddi_dma_handle_t	dma_handle;
    uint8_t		dma_bound;
    size_t		reallen;
    ddi_acc_handle_t	acc_handle;

    struct bdl_chunk   *next;		/* next DMA data chunk, when on free list */
};


/*
 * 
 */
struct sample_buffer {
    uint8_t		io_started;
    uint8_t		head;		/* ICH audio controller accesses sample */
    					/* buffer starting with this index */
    uint8_t		tail;		/* first free slot after the end of valid  */
    					/* buffers */
    uint8_t		buffers_available;
    struct bdl_chunk   *active[I810_NUM_BD];
					/* while chunks are in use by the ICH audio
					 * controller, they are placed in the active 
					 * array under their current bdl slot index
					 */
    struct bdl_chunk   *free;		/* list of unused sample chunks */
};
typedef struct sample_buffer sample_buffer_t;


/*
 * ICH buffer descriptor list for DMA
 * From 2.2.1 of the ICH programmers manual.
 */

struct i810_buf_descr {
    uint32_t	bd_ptr;		/* location of the data buffer (must be an even addr) */
    uint16_t    bd_len:16;	/* length of buffer, in samples */
    uint16_t	:14;
    uint8_t	bd_bup:1;	/* buffer underrun policy */
    uint8_t	bd_ioc:1;	/* interrupt on completion flag */
};
typedef struct i810_buf_descr i810_buf_descr_t;


/*
 * We allocate the 32-entry buffer descriptor lists for each 
 * of the three DMA engines available in the ICH audio controller
 * in one single 4k physical page:
 */
struct i810_bdl {
    i810_buf_descr_t	pcm_in_bdl[I810_NUM_BD];
    i810_buf_descr_t	pcm_out_bdl[I810_NUM_BD];
    i810_buf_descr_t	mic_in_bdl[I810_NUM_BD];
};
typedef struct i810_bdl i810_bdl_t;



/*
 * Per device softstate
 */
typedef struct audioi810_softstate {
    dev_info_t         *dip;            /* device info node for your sound hw */
    audiohdl_t		audio_handle;	/* opaque handle to audio support module */

    audio_codec_handle_t codec_handle;

    kmutex_t            instance_mutex;
    kmutex_t            intr_mutex;
    ddi_iblock_cookie_t	intr_cookie;

    caddr_t		am_regs;	/* AC97 audio I/O space registers */
    ddi_acc_handle_t	am_regs_handle;
    caddr_t		bm_regs;	/* ICH bus master I/O space registers */
    ddi_acc_handle_t	bm_regs_handle;

    uint8_t		bm_sr_reg;	/* Status reg offset, ICH: 6 SiS: 8 */
    uint8_t		bm_picb_reg;	/* PICB   reg offset, ICH: 8 SiS: 6 */
    uint8_t		sample_size;	/* ICH: 2, SiS: 1 */

    uint8_t		power_down;
    uint8_t		saved_iline;

    ddi_acc_handle_t	pci_conf_handle;

    char               *revstr;         /* pci revision of card as printable string */
    uint8_t             rev;            /* pci revision byte of card */
    uint8_t		ich4;		/* 1: Intel ICH4 */
    uint8_t		sis;		/* 1: SiS 7012, 2: 7012 + large bm reg block */
    uint8_t		codec;

    uint16_t		ac97_ext_features;
    uint16_t		ac97_clock;	/* internal clock rate of AC'97 codec */

    ddi_dma_handle_t	bdl_dma_handle;	/* DMA handle for buffer descriptor list */
    int			bdl_dma_bound;
    ddi_acc_handle_t	bdl_acc_handle;
    i810_bdl_t	       *bdl;		/* virtual kernel addr of bdl */
    i810_bdl_t	       *bdl_paddr;	/* physical   	    "	      */
    size_t		bdllen;

    sample_buffer_t	record_buf;
    sample_buffer_t	play_buf;
    timeout_id_t	play_timeout_id;

    audio_info_t	audio_info;	/* driver local audio state. Note: */
    					/* we maintain only a few fields */
    					/* of the structure in the driver */

    audio_device_t	dev_info;
    ah_ad_info_t	ad_info;
    audio_info_t	ad_defaults;
} audioi810_softstate_t;


#define	TRUE	1
#define	FALSE	0

/* general success/failure return codes */
#define	AUDIOI810_SUCCESS	0
#define	AUDIOI810_FAILURE	(-1)


/*
 * macros to access the ICH native audio mixer registers 
 * These registers are all 16-bit registers.
 */
#define	I810_AM_GET16(offs)					\
	ddi_get16(softstate->am_regs_handle,			\
		  (void*) ((char*)softstate->am_regs + (offs)))
#define	I810_AM_PUT16(offs, val)				\
	ddi_put16(softstate->am_regs_handle,			\
		  (void*) ((char*)softstate->am_regs + (offs)),	\
		  (val))

#define	I810_AM_REGS_SIZE	0x100	/* NAMBAR: size of audio mixer register block */
#define	I845_AM_REGS_SIZE	0x200	/* ICH4: MMBAR audio mixer register block */


/*
 * macros to access the ICH native audio bus mastering registers 
 */
#define	I810_BM_GET8(offs)					\
	ddi_get8(softstate->bm_regs_handle,			\
		  (void*) ((char*)softstate->bm_regs + (offs)))
#define	I810_BM_GET16(offs)					\
	ddi_get16(softstate->bm_regs_handle,			\
		  (void*) ((char*)softstate->bm_regs + (offs)))
#define	I810_BM_GET32(offs)					\
	ddi_get32(softstate->bm_regs_handle,			\
		  (void*) ((char*)softstate->bm_regs + (offs)))

#define	I810_BM_PUT8(offs, val)					\
	ddi_put8(softstate->bm_regs_handle,			\
		  (void*) ((char*)softstate->bm_regs + (offs)),	\
		  (val))
#define	I810_BM_PUT16(offs, val)				\
	ddi_put16(softstate->bm_regs_handle,			\
		  (void*) ((char*)softstate->bm_regs + (offs)),	\
		  (val))
#define	I810_BM_PUT32(offs, val)				\
	ddi_put32(softstate->bm_regs_handle,			\
		  (void*) ((char*)softstate->bm_regs + (offs)),	\
		  (val))


  /* CODEC2 is for stuff like "surround sound". We wont be using it */
#define I810_CODEC_OFFSET	0x80
  
/* Offsets for SECOND register set.
 * Table 13-3 in ICH2-M datasheet
 */
  
#define I810_BM_DESCLIST	0x00
#define I810_BM_CIV		0x04	/* Current index value */
#define I810_BM_LVI		0x05	/* Last valid index */
#define I810_BM_SR		0x06	/* Status register... */
					/* can be read in 4-byte from PI_CIV*/
#define I810_BM_PICB		0x08	/* Position in current buffer */
#define I810_BM_PIV		0x0a	/* Prefetched index value */
#define I810_BM_CR		0x0b /*1byte*/ /* control register */

/*
 * SiS 7012: has SR & PICB registers swapped
 */
#define SIS_BM_SR		0x08	/* Status register... */
#define SIS_BM_PICB		0x06	/* Position in current buffer */

/* _BusMaster_PCMIn */
#define I810_BM_PI_OFFSET	0x00
#define	I810_BM_PI_DESCLIST	(I810_BM_PI_OFFSET + I810_BM_DESCLIST)
#define I810_BM_PI_CIV		(I810_BM_PI_OFFSET + I810_BM_CIV)
#define I810_BM_PI_LVI		(I810_BM_PI_OFFSET + I810_BM_LVI)
#define I810_BM_PI_SR		(I810_BM_PI_OFFSET + I810_BM_SR)
#define I810_BM_PI_PICB		(I810_BM_PI_OFFSET + I810_BM_PICB)
#define I810_BM_PI_PIV		(I810_BM_PI_OFFSET + I810_BM_PIV)
#define I810_BM_PI_CR		(I810_BM_PI_OFFSET + I810_BM_CR)
/*_BusMaster_PCMOut */
#define I810_BM_PO_OFFSET	0x10
#define	I810_BM_PO_DESCLIST	(I810_BM_PO_OFFSET + I810_BM_DESCLIST)
#define I810_BM_PO_CIV		(I810_BM_PO_OFFSET + I810_BM_CIV)
#define I810_BM_PO_LVI		(I810_BM_PO_OFFSET + I810_BM_LVI)
#define I810_BM_PO_SR		(I810_BM_PO_OFFSET + I810_BM_SR)
#define I810_BM_PO_PICB		(I810_BM_PO_OFFSET + I810_BM_PICB)
#define I810_BM_PO_PIV		(I810_BM_PO_OFFSET + I810_BM_PIV)
#define I810_BM_PO_CR		(I810_BM_PO_OFFSET + I810_BM_CR)
/*_BusMaster_Microphone(in) */
#define I810_BM_MC_OFFSET	0x20
#define	I810_BM_MC_DESCLIST	(I810_BM_MC_OFFSET + I810_BM_DESCLIST)
#define I810_BM_MC_CIV		(I810_BM_MC_OFFSET + I810_BM_CIV)
#define I810_BM_MC_LVI		(I810_BM_MC_OFFSET + I810_BM_LVI)
#define I810_BM_MC_SR		(I810_BM_MC_OFFSET + I810_BM_SR)
#define I810_BM_MC_PICB		(I810_BM_MC_OFFSET + I810_BM_PICB)
#define I810_BM_MC_PIV		(I810_BM_MC_OFFSET + I810_BM_PIV)
#define I810_BM_MC_CR		(I810_BM_MC_OFFSET + I810_BM_CR)
/*
 * The following three are shared with the "windows-modem" hardware on the chip.
 */
#define I810_GLOBAL_CONTROL	0x2c /*4byte*/
#define I810_GLOBAL_STATUS	0x30 /*4byte*/
#define I810_CODEC_SEMA		0x34 /*1byte*/

#define	I810_BM_REGS_SIZE	0x40	/* NABMBAR: size of audio bus mastering register block */
#define	SIS_BM_REGS_SIZE	0x80	/* SiS7012 has extended the I810 NABMBAR block */
#define	I845_BM_REGS_SIZE	0x100	/* ICH4: MBBAR size of audio bus mastering register block */

/* CODEC_SEMA(PHORE) is "special". It is actually a semaphore of sorts.
   (Really, more of an auto-resetting mutex)
  Read CODEC_SEMA, and IFF it is 0, you can safely do a "codec access".
  If 1, then loop until 0.
  The act of reading it will set it to 1 AFTER you have read it.
  Once you have "accessed the codec", it will reset to 0.
  UGH! This sounds like we're basically doing "port" style access.
  And essentially we are, since this is going over a hidden ICH "hub"
  
  
*/
  

/************************************************************/
/* Defines for bits on busmaster registers                  */
/************************************************************/
 /* Any of the I810_BM_XX_SR regs use these */
#define I810_BM_STATUS_FIFOERR		0x10
#define I810_BM_STATUS_BUFCOMPLETE	0x08 /* finished _A_ buffer*/
#define I810_BM_STATUS_LASTBUFCOMPLETE	0x04 /* finished _ALL_ buffers */
#define I810_BM_STATUS_ONLASTBUF	0x02
#define I810_BM_STATUS_DMAHALTED	0x01

 /* Any of the I810_BM_XX_CR regs use these */
#define I810_BM_CONTROL_IOCENABLE	0x10
#define I810_BM_CONTROL_FEIENABLE	0x08
#define I810_BM_CONTROL_LVBENABLE	0x04 /* enable interrupt when DMAdone*/
#define I810_BM_CONTROL_RESETREGS	0x02
#define I810_BM_CONTROL_DMASTART	0x01 /* 0 to pause, 1 to run normal*/

/* for I810_GLOBAL_CONTROL reg */
#define I810_GLBCTRL_2CHANNEL		(0<<20)	/* ICH2: PCM 4/6 Enable */
#define I810_GLBCTRL_4CHANNEL		(1<<20)
#define I810_GLBCTRL_6CHANNEL		(2<<20)
#define I810_GLBCTRL_CHANNEL_MASK	(3<<20)	/* ICH2: PCM 4/6 Enable */
#define I810_GLBCTRL_RESUME2		0x20
#define I810_GLBCTRL_RESUME		0x10
#define I810_GLBCTRL_ACLINKOFF		0x08 /* turns off "AC Link" (?) */
#define I810_GLBCTRL_WARMRESET		0x04 /* write bit to reset */
					     /* blk_clk must be first stopped*/
#define I810_GLBCTRL_COLDRESET		0x02 /* write 0, then 1, to reset*/
#define I810_GLBCTRL_GIE		0x01 /* enable interrupt on change
					      * of any "GPI" ??
					      */

/* for I810_GLOBAL_STATUS reg */
#define I810_GLBSTAT_CODEC3RDY	0x10000000 /* 3rd codec ready */
#define I810_GLBSTAT_SUPP6	0x00200000
#define I810_GLBSTAT_SUPP4	0x00100000
#define I810_GLBSTAT_CODECTMOUT	0x00008000
#define I810_GLBSTAT_CODEC2RDY	0x00000200 /* 2nd codec ready */
#define I810_GLBSTAT_CODECRDY	0x00000100 /* 1st codec ready */
#define I810_GLBSTAT_MICIRQ	0x00000080 /* autoclear */
#define I810_GLBSTAT_POUTIRQ	0x00000040 /* autoclear */
#define I810_GLBSTAT_PINIRQ	0x00000020 /* autoclear */

#endif  /* _KERNEL */


/* custom 'ioctl' supported by the audioi810 driver */
#define	AUDIOI810_IOC			('i'<<8)
#define	AUDIOI810_SET_PCMOUT_VOL	(AUDIOI810_IOC | 0)
#define	AUDIOI810_SET_BEEP_VOL		(AUDIOI810_IOC | 1)
#define	AUDIOI810_SET_LOUDNESS_ENABLE	(AUDIOI810_IOC | 2)
#define	AUDIOI810_SET_3D_STEREO_ENABLE	(AUDIOI810_IOC | 3)
#define	AUDIOI810_SET_3D_STEREO_DEPTH	(AUDIOI810_IOC | 4)
#define	AUDIOI810_SET_3D_STEREO_CENTER	(AUDIOI810_IOC | 5)
#define	AUDIOI810_SET_STEREO_ENHANCE_ENABLE	(AUDIOI810_IOC | 6)
