/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/note.h>
#include <sys/modctl.h>
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/sunddi.h>
#include <sys/cmn_err.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/mixer.h>
#include <sys/audio.h>
#include <sys/audio/audio_helper.h>
#include <sys/ddi.h>

#include "audiohlp.h"


/*
 * Module linkage information for the kernel.
 */

static struct modlmisc modlmisc = {
    &mod_miscops,     /* Type of module.  This one is a misc module */
    "audio helper module" 
};

static struct modlinkage modlinkage = {
    MODREV_1,
    { &modlmisc, NULL },
};


uint_t audiohlp_errlevel = AUDL_L2;
uint_t audiohlp_errmask  = AUDM_ALL;

void
audiohlp_vprintf(uint_t level, uint_t mask, char *fmt, va_list args)
{
    if (level <= audiohlp_errlevel && (mask & audiohlp_errmask))
	vcmn_err(CE_CONT, fmt, args);
}


#ifdef	DEBUG
void
audiohlp_dprintf4(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiohlp_vprintf(AUDL_L4, mask, fmt, args);
    va_end(args);
}

void
audiohlp_dprintf3(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiohlp_vprintf(AUDL_L3, mask, fmt, args);
    va_end(args);
}

void
audiohlp_dprintf2(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiohlp_vprintf(AUDL_L2, mask, fmt, args);
    va_end(args);
}
#endif

void
audiohlp_dprintf1(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiohlp_vprintf(AUDL_L1, mask, fmt, args);
    va_end(args);
}

void
audiohlp_dprintf0(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiohlp_vprintf(AUDL_L0, mask, fmt, args);
    va_end(args);
}


static kmutex_t audiohlp_state_lock;

int
_init(void)
{
    int err;

    AUD_DPRINTF_L4(AUDM_MODS, "_init\n");

    mutex_init(&audiohlp_state_lock, NULL, MUTEX_DRIVER, NULL);

    if ((err = mod_install(&modlinkage)) != 0)
	mutex_destroy(&audiohlp_state_lock);
    return err;
}

int
_fini(void)
{
    int err;

    AUD_DPRINTF_L4(AUDM_MODS, "_fini\n");
    
    if ((err = mod_remove(&modlinkage)) == 0)
	mutex_destroy(&audiohlp_state_lock);
    return err;
}

int
_info(struct modinfo *modinfop)
{
    AUD_DPRINTF_L4(AUDM_MODS, "_info\n");
    return mod_info(&modlinkage, modinfop);
}


#ifdef	SOLARIS9
/*
 * solaris 9 has ddi_driver_major()
 */
#else
/*
 * reimplement ddi_driver_major()
 */
#define ddi_driver_major(dip)	get_devinfo_major(dip)

#define	NOMAJOR	((major_t)-1)

static major_t
get_devinfo_major(dev_info_t *dip)
{
    char *name;

    if ((name = ddi_binding_name(dip)) == NULL)
	return NOMAJOR;

    return ddi_name_to_major(name);
}
#endif


struct device_state {
    struct device_state	*next;
    major_t		 major;
    int			 item;
    void		*state;
};
struct device_state *audio_hlp_soft_states;


int
audio_hlp_soft_state_zalloc(major_t major, int item)
{
    struct device_state *s, *new;

    if ((new = kmem_alloc(sizeof(struct device_state), KM_SLEEP)) == NULL)
	return DDI_FAILURE;
    if ((new->state = kmem_zalloc(sizeof(struct audio_handle), KM_SLEEP)) == NULL) {
	kmem_free(new, sizeof(struct device_state));
	return DDI_FAILURE;
    }

    mutex_enter(&audiohlp_state_lock);
    for (s = audio_hlp_soft_states; s != NULL; s = s->next) {
	if (s->major == major && s->item == item) {
	    mutex_exit(&audiohlp_state_lock);
	    kmem_free(new->state, sizeof(struct audio_handle));
	    kmem_free(new, sizeof(struct device_state));
	    return DDI_FAILURE;
	}
    }
    new->next = audio_hlp_soft_states;
    audio_hlp_soft_states = new;
    new->major = major;
    new->item = item;
    mutex_exit(&audiohlp_state_lock);

    return DDI_SUCCESS;
}


void *
audio_hlp_get_soft_state(major_t major, int item)
{
    struct device_state *s;

    mutex_enter(&audiohlp_state_lock);
    for (s = audio_hlp_soft_states; s != NULL; s = s->next) {
	if (s->major == major && s->item == item) {
    	    mutex_exit(&audiohlp_state_lock);
    	    return s->state;
	}
    }
    mutex_exit(&audiohlp_state_lock);
    return NULL;
}


void
audio_hlp_soft_state_free(major_t major, int item)
{
    struct device_state *s, *prev;

    mutex_enter(&audiohlp_state_lock);
    for (s = audio_hlp_soft_states, prev = NULL;
	 s != NULL;
	 prev = s, s = s->next) {

	if (s->major == major && s->item == item) {
	    if (prev)
		prev->next = s->next;
	    else
		audio_hlp_soft_states = s->next;
	    kmem_free(s->state, sizeof(struct audio_handle));
	    kmem_free(s, sizeof(struct device_state));
	    break;
	}
    }
    mutex_exit(&audiohlp_state_lock);
}


/*
 * Applications that open the audio control pseudo-device may request
 * asynchronous notification of changes in the state of the audio
 * device by setting the S_MSG flag in an I_SETSIG STREAMS ioctl. Such
 * processes receive a SIGPOLL signal when certain events occur.
 *
 * Send such a SIGPOLL notification signal to interested processes.
 */
void
audio_hlp_status_changed(struct audio_handle *handle)
{
    queue_t *q;
    mblk_t *mp;

    /* any audioctl device open? */
    if ((q = handle->ctl_q) == NULL)
	return;

    AUD_DPRINTF_L4(AUDM_ALL, "send SIGPOLL notification");

    mp = allocb(1, BPRI_HI);
    if (mp == NULL) {
	cmn_err(CE_NOTE, "no mem for status changed msg");
	return;
    }
    
    mp->b_datap->db_type = M_PCSIG;
    *mp->b_wptr++ = SIGPOLL;

    /* send M_PCSIG message with a SIGPOLL signal to the audioctl read queue */
    putnext(q, mp);
}


/*
 * Wait until all playback buffers have played or a signal is
 * delivered to the process.
 */
void
audio_hlp_audio_drain(queue_t *q)
{
    audiohdl_t handle = audio_hlp_get_qptr_data(q);

    mutex_enter(&handle->lock);

    handle->drain_waiting = 1;

    while (qsize(handle->play_q) > 0 || handle->audio_info.play.active) {
	int rv;

	AUD_DPRINTF_L4(AUDM_ALL, "waiting for playback end qsize %d, act %d",
		       qsize(handle->play_q), handle->audio_info.play.active);

#if	1
	{
	    audio_prinfo_t *prinfo = &handle->audio_info.play;
	    /* 
	     * during driver development, we also terminate after a
	     * reasonable timeout
	     */
	    int pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
			 (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
			  prinfo->encoding != AUDIO_ENCODING_LINEAR8));
	    int stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);
	    
	    int delay =
		  (q->q_hiwat /
		   (prinfo->precision == AUDIO_PRECISION_16 ? 2 : 1) )
		+ (handle->ah_ad_info->ad_defaults->play.buffer_size /
		   (pcm16 ? 2 : 1) );

	    delay = 1 +			/* 1 second minimum delay */
		(3 * delay / 2		/* + 50% safety margin */
		 / prinfo->sample_rate
		 / (stereo ? 2 : 1));
		 
	    rv = cv_timedwait_sig(&handle->playback_empty_cv,
				  &handle->lock,
				  ddi_get_lbolt() +
				  delay * drv_usectohz(1000000));
	}
#else
	rv = cv_wait_sig(&handle->playback_empty_cv,
			 &handle->lock);
#endif
	if (rv == -1) {
	    cmn_err(CE_WARN, "timeout waiting for playback end");
	    break;
	} else if (rv == 0) {
	    AUD_DPRINTF_L4(AUDM_ALL, "interrupted waiting for playback end");
	    break;
	}
    }

    handle->drain_waiting = 0;

    mutex_exit(&handle->lock);
}


/*
 * Audio Support Module Entry Point Routines
 */
audiohdl_t
audio_hlp_attach(dev_info_t *dip, ddi_attach_cmd_t cmd,
		 struct ah_ad_info *ah_ad_info)
{
    major_t major = ddi_driver_major(dip);
    int instance = ddi_get_instance(dip);
    struct audio_handle *handle;

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_attach dev %d, inst %d\n",
		   major, instance);

    switch (cmd) {
    case DDI_ATTACH:
	/* we handle the ATTACH command */
	break;

    default:
	/* everything else is an error */
	return NULL;
    }

    if (ah_ad_info->ad_int_vers != AD_VERSION) {
	cmn_err(CE_NOTE, "bad audio driver interface version %d",
		ah_ad_info->ad_int_vers);
	return NULL;
    }

    if (audio_hlp_soft_state_zalloc(major, instance) != DDI_SUCCESS)
	return NULL;

    if ((handle = audio_hlp_get_soft_state(major, instance)) == NULL)
	return NULL;

    handle->ah_dip = dip;

    handle->ah_ad_info = ah_ad_info;

    mutex_init(&handle->lock, NULL, MUTEX_DRIVER, NULL);
    cv_init(&handle->close_wait_cv, NULL, CV_DRIVER, NULL);
    cv_init(&handle->playback_empty_cv, NULL, CV_DRIVER, NULL);

    /*
     * It seems there are hardcoded minor device names in
     * /usr/lib/devfsadm/linkmod/SUNW_audio_link.so,
     * so we cannot use the minor node name "sbpci" or
     * "sound,sbpci".
     *
     * It seems the SUNW_audio_link.so library knows about
     * sound,audio and sound,sbpro...
     *
     */

    ddi_create_minor_node(dip, "sound,audio", S_IFCHR,
			  AUDIOHLP_MAKEDEV(instance, AUDIOHLP_AUDIO_MIN),
			  DDI_NT_AUDIO, 0);
	
    ddi_create_minor_node(dip, "sound,audioctl", S_IFCHR,
			  AUDIOHLP_MAKEDEV(instance, AUDIOHLP_AUDIOCTL_MIN),
			  DDI_NT_AUDIO, 0);

    return handle;
}


int
audio_hlp_detach(audiohdl_t handle, ddi_detach_cmd_t cmd)
{
    major_t major;
    int instance;

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_detach\n");

    switch (cmd) {
    case DDI_DETACH:
	break;

    default:
	return DDI_FAILURE;
    }

    if (handle == NULL) {
	cmn_err(CE_NOTE, "audio_hlp_detach: bad handle");
	return DDI_FAILURE;
    }

    major = ddi_driver_major(handle->ah_dip);
    instance = ddi_get_instance(handle->ah_dip);
    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_detach dev %d, inst %d\n",
		   major, instance);

    ddi_remove_minor_node(handle->ah_dip, NULL);

    mutex_destroy(&handle->lock);
    cv_destroy(&handle->close_wait_cv);
    cv_destroy(&handle->playback_empty_cv);

    audio_hlp_soft_state_free(major, instance);

    return DDI_SUCCESS;
}


/*
 * Initialize audio state info for the device to default
 * parameters.
 */
static void
audio_hlp_default_audioparams(audiohdl_t handle)
{
    audio_info_t oldai;
    audio_info_t *audio_info = &handle->audio_info;
    audio_prinfo_t *play, *rec;

    /*
     * save a copy of the old values, we need restore certain values
     * if this is not the first time initialization
     */
    bcopy(audio_info, &oldai, sizeof(audio_info_t));

    bcopy(handle->ah_ad_info->ad_defaults, audio_info, sizeof(audio_info_t));

    play = &audio_info->play;
    rec = &audio_info->record;

    if (handle->initialized) {
	/*
	 * several parameters persist through subsequent opens:
	 * {play,record}.{gain,balance,port}
	 * monitor_gain
	 * output_muted
	 */
	play->gain    = oldai.play.gain;
	play->balance = oldai.play.balance;
	play->port    = oldai.play.port;

	rec->gain     = oldai.record.gain;
	rec->balance  = oldai.record.balance;
	rec->port     = oldai.record.port;

	audio_info->monitor_gain = oldai.monitor_gain;
	audio_info->output_muted = oldai.output_muted;
    }

    handle->initialized = TRUE;

    audio_hlp_audio_setdefaults(handle);
}



int
audio_hlp_open(queue_t *q, dev_t *devp, int flag, int sflags,
	       cred_t *credp)
{
    int instance = AUDIOHLP_INST(*devp);
    audiohdl_t handle;
#if     DEBUG
    pid_t open_pid = ddi_get_pid();
#endif
    int i;

    AUD_DPRINTF_L4(AUDM_OPEN, "audio_hlp_open\n");

    if (sflags) {	/* audio driver refuses to do module or clone open */
        return ENXIO;
    }

    handle = audio_hlp_get_soft_state(getmajor(*devp), instance);
    if (handle == NULL) {
	return ENXIO;
    }

    mutex_enter(&handle->lock);

    if (handle->openflags == 0) {
	if (AD_SETUP(handle, AUDIO_NO_CHANNEL, 0) != AUDIO_SUCCESS) {
            mutex_exit(&handle->lock);
            return ENXIO;
        }
    }

    if ((handle->openflags & ~(1<<AUDIOHLP_AUDIOCTL_MIN)) == 0) {
        /*
         * All minor devices for this instance were closed up to now,
	 * ignoring audioctl devices.
         *
         * Thus configure audio device for default audio parameters:
         * 8kHz, 8-bit u-Law, mono
         */
        AUD_DPRINTF_L4(AUDM_OPEN, "using default audio format: 8kHz, 8-bit u-Law, mono");

	audio_hlp_default_audioparams(handle);
    }
    mutex_exit(&handle->lock);

    switch (AUDIOHLP_MINOR(*devp)) {
    default:
        AUD_DPRINTF_L4(AUDM_OPEN, "direct driver open on clone device");
        return ENXIO;

    case AUDIOHLP_AUDIOCTL_MIN:
        /*
         * An open on the /dev/audioctl device is always allowed.
         *
         * Remember the STREAMS queue for audio status notification events.
         */
        handle->ctl_q = q;
        break;

    case AUDIOHLP_AUDIO_MIN:
        /*
         * Clone open for the /dev/audio device.
         *
         * Only one process is granted read or write access at one time.
         * If the device is already open by another process, the open
         * blocks, unless the O_NONBLOCK or O_NDELAY flag was passed;
         * the open returns with error EBUSY in that case.
         */
        mutex_enter(&handle->lock);

        while (((flag & FREAD)  && handle->audio_info.record.open) ||
               ((flag & FWRITE) && handle->audio_info.play.open)) {

            /*
             * The device is busy for the requested access mode
             * by some other process. Decide if we return an error
             * or block until the device is released.
             */

            if (flag & (FNONBLOCK|FNDELAY)) {
                mutex_exit(&handle->lock);
                return EBUSY;
            }

            AUD_DPRINTF_L4(AUDM_OPEN, "open blocked: record=%d, play=%d, pid=%ld",
		    handle->audio_info.record.open,
		    handle->audio_info.play.open,
		    (long)open_pid);

            if (flag & FREAD)
                handle->audio_info.record.waiting = 1;
            if (flag & FWRITE)
                handle->audio_info.play.waiting   = 1;
            audio_hlp_status_changed(handle);

            /* wait until someone closes an audio device */
            if (cv_wait_sig(&handle->close_wait_cv,
                            &handle->lock) == 0) {

                AUD_DPRINTF_L4(AUDM_OPEN, "open wait interrupted");
                mutex_exit(&handle->lock);
                return EINTR;
            }
        }

        /* pick an unused minor device for the clone open */
        for (i = 0; i < AUDIOHLP_CLONE_COUNT; i++) {
            if (!handle->minor_inuse[i])
                break;
        }
        if (i >= AUDIOHLP_CLONE_COUNT) {
                mutex_exit(&handle->lock);
                return ENXIO;
        }
        if  (flag & FWRITE) handle->minor_inuse[i] |= AUDIO_PLAY;
        if  (flag & FREAD)  handle->minor_inuse[i] |= AUDIO_RECORD;

        /* clone it */
        *devp = makedevice(getmajor(*devp),
			   AUDIOHLP_MAKEDEV(AUDIOHLP_INST(*devp),
					    AUDIOHLP_CLONE_MIN + i));


        if (flag & FREAD) {
            handle->audio_info.record.open     = 1;
            handle->audio_info.record.waiting  = 0;
            handle->audio_info.record.eof      = 0;
            handle->audio_info.record.samples  = 0;
            handle->audio_info.record.error    = 0;
            handle->audio_info.record.pause    = 0;
	    handle->audio_info.record.active   = TRUE;
            handle->audio_info.record.minordev = AUDIOHLP_CLONE_MIN + i;
	    handle->record_q = q;
        }
        if (flag & FWRITE) {
            handle->audio_info.play.open     = 1;
            handle->audio_info.play.waiting  = 0;
            handle->audio_info.play.eof      = 0;
            handle->audio_info.play.samples  = 0;
            handle->audio_info.play.error    = 0;
            handle->audio_info.play.pause    = 0;
	    handle->audio_info.play.active   = FALSE;
            handle->audio_info.play.minordev = AUDIOHLP_CLONE_MIN + i;
	    handle->play_q = WR(q);
        }
#ifdef  AUDIOIO_HAS_REFCNT
        handle->audio_info.ref_cnt++;
#endif
        audio_hlp_status_changed(handle);

        mutex_exit(&handle->lock);

        AUD_DPRINTF_L4(AUDM_OPEN, "clone open, *dev=%lx, pid=%ld",
		       (long)*devp, (long)open_pid);
        break;
    }

    audio_hlp_set_qptr(q, *devp, handle);

    mutex_enter(&handle->lock);
    handle->openflags |= 1 << AUDIOHLP_MINOR(*devp);
    mutex_exit(&handle->lock);

    qprocson(q);

    if (!AUDIOHLP_ISCTL(*devp)) {
        if (flag & FWRITE) {
	    noenable(q);
        }
        if (flag & FREAD) {
            /* start recording of samples */
            AD_START_RECORD(handle, AUDIO_NO_CHANNEL);
        }
    }

    return 0;
}


int
audio_hlp_close(queue_t *q, int flag, cred_t *credp)
{
    audiohdl_t handle = audio_hlp_get_qptr_data(q);
    int minor;

    AUD_DPRINTF_L4(AUDM_CLOSE, "audio_hlp_close\n");

    if (handle == NULL)
	return ENXIO;

    switch (minor = AUDIOHLP_MINOR(audio_hlp_get_qptr_dev(q))) {
    case AUDIOHLP_AUDIO_MIN:
        /* should not happen, we clone opened in that case */
        break;

    case AUDIOHLP_AUDIOCTL_MIN:
        /*
         * last close on /dev/audioctl, stop sending audio
         * status notifications
         */
        handle->ctl_q = NULL;
        break;

    default:
        /*
         * close a cloned /dev/audio device
         */
        if (minor <  AUDIOHLP_CLONE_MIN ||
            minor >= AUDIOHLP_CLONE_MIN + AUDIOHLP_CLONE_COUNT) {

            AUD_DPRINTF_L4(AUDM_CLOSE, "audio_hlp close: invalid clone device");
            return EINVAL;
        }

        if (flag & FWRITE) {
            audio_hlp_audio_drain(q);
	    AD_STOP_PLAY(handle, AUDIO_NO_CHANNEL);
	    handle->play_q = NULL;

	    if (handle->play_mp) {
		AUD_DPRINTF_L2(AUDM_CLOSE, "audio_hlp close: play_mp != NULL");
		/*
		freemsg(handle->play_mp);
		handle->play_mp = 0;
		*/
	    }
        }
        if (flag & FREAD) {
            AD_STOP_RECORD(handle, AUDIO_NO_CHANNEL);
	    handle->record_q = NULL;
        }

        mutex_enter(&handle->lock);

        if (flag & FREAD) {
            handle->audio_info.record.open     = 0;
            handle->audio_info.record.waiting  = 0;
	    handle->audio_info.record.active   = FALSE;
            handle->audio_info.record.minordev = ~0;
        }
        if (flag & FWRITE) {
            handle->audio_info.play.open     = 0;
            handle->audio_info.play.waiting  = 0;
            handle->audio_info.play.minordev = ~0;
        }
#ifdef  AUDIOIO_HAS_REFCNT
        handle->audio_info.ref_cnt--;
#endif
        audio_hlp_status_changed(handle);

        handle->minor_inuse[minor - AUDIOHLP_CLONE_MIN] = 0;

        cv_signal(&handle->close_wait_cv);

        mutex_exit(&handle->lock);

        break;
    }
    /*
     * if this close was the last user of the device,
     * reset the hardware to some default state.
     */
    mutex_enter(&handle->lock);

    handle->openflags &= ~ (1 << AUDIOHLP_MINOR(audio_hlp_get_qptr_dev(q)));
    if (handle->openflags == 0) {

        AD_TEARDOWN(handle, AUDIO_NO_CHANNEL, 0);

    }
    mutex_exit(&handle->lock);

    qprocsoff(q);

    audio_hlp_free_qptr(q);

    return 0;    
}


int
audio_hlp_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
		  void *arg, void **result)
{
    audiohdl_t handle;
    int instance;
    int error = DDI_FAILURE;

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_getinfo, dip=%p, infocmd=%d\n", dip, infocmd);

    switch (infocmd) {
    case DDI_INFO_DEVT2DEVINFO:
	instance = AUDIOHLP_INST((dev_t) arg);
	if ((handle = audio_hlp_get_soft_state(getmajor((dev_t) arg),
					       instance)) != NULL) {
	    *result = handle->ah_dip;
	    error = DDI_SUCCESS;
	} else
	    *result = NULL;
	break;
	
    case DDI_INFO_DEVT2INSTANCE:
	instance = AUDIOHLP_INST((dev_t) arg);
	*result = (void *)(long)instance;
	error = DDI_SUCCESS;
	break;
    }

    return error;
}


static char *
mblk_type(mblk_t *mp)
{
    switch (mp->b_datap->db_type) {
    case M_DATA:        return "M_DATA";
    case M_PROTO:       return "M_PROTO";
    case M_IOCTL:       return "M_IOCTL";
    case M_FLUSH:       return "M_FLUSH";
    case M_IOCDATA:     return "M_IOCDATA";
    default:            return "???";
    }
}


int
audio_hlp_rput(queue_t *q, mblk_t *mp)
{
    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_rput\n");
    putq(q, mp);
    return 0;
}


/*
 * STREAMS upstream (read) service procedure.
 */
int
audio_hlp_rsvc(queue_t *q)
{
    mblk_t *mp;
    int free_space = FALSE;

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_rsvc\n");

    while ((mp = getq(q)) != NULL) {
	if (!canputnext(q)) {
	    /*
	     * cannot send upstream, putback on our queue for a later
	     * rsvc retry
	     */
	    if (!putbq(q, mp)) {
		/*
		 * too bad, cannot place the message back on the
		 * queue.  Don't leak the message.
		 * XXX: should we signal a record overflow error?
		 */
		freemsg(mp);
	    }
	    break;
	}
	putnext(q, mp);
	free_space = TRUE;
    }

    if (free_space) {
    }
    return 0;
}


/*
 * STREAMS downstream (write) put procedure.
 */
int
audio_hlp_wput(queue_t *q, mblk_t *mp)
{
    audiohdl_t handle = audio_hlp_get_qptr_data(q);
    struct copyresp *copyresp;
    struct iocblk *iocp;
    int ret, error;

    AUD_DPRINTF_L4(AUDM_STRMSG, "audio_hlp_wput, %s(%x), sz %ld\n",
		   mblk_type(mp), mp->b_datap->db_type,
		   (long)(mp->b_wptr - mp->b_rptr));

    switch (mp->b_datap->db_type) {
    case M_IOCTL:
	error = 0;
	ret = audio_hlp_ioctl(handle, q, mp, &error);
	switch (ret) {
	case AD_WIOCDATA:
	    /*
	     * caller is responsible for modifying mp to the desired
	     * M_COPYIN/M_COPYOUT message
	     */
	    break;

	default:
	    error = 0;
	case AD_ACK:
	case AD_NACK:
	    iocp = (struct iocblk *)mp->b_rptr;
	    iocp->ioc_count = 0;
	    iocp->ioc_rval = 0;
	    iocp->ioc_error = error;
	    mp->b_datap->db_type = ret == AD_ACK ? M_IOCACK :  M_IOCNAK; 
	    mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
	    if (mp->b_cont) {
		freemsg(mp->b_cont);
		mp->b_cont = NULL;
	    }
	    break;
	}
	qreply(q, mp);
	break;

    case M_IOCDATA:
	copyresp = (struct copyresp *)mp->b_rptr;
	
	if (copyresp->cp_rval != 0) {
	    /* the copyin/copyout failed; we fail the ioctl, too */
	    iocp = (struct iocblk *)mp->b_rptr;
	    iocp->ioc_count = 0;
	    iocp->ioc_error = EINVAL;
	    mp->b_datap->db_type = M_IOCNAK;
	    mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
	    /* send response to the original M_IOCTL */
	    qreply(q, mp);
	    break;
 	}

	if (mp->b_cont == NULL) {
	    /*
	     * a response to a M_COPYOUT message.  Send back the
	     * final M_IOCACK to finish the whole M_IOCTL req.
	     */
	    iocp = (struct iocblk *)mp->b_rptr;
	    iocp->ioc_count = 0;
	    iocp->ioc_error = 0;
	    mp->b_datap->db_type = M_IOCACK;
	    mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
	    qreply(q, mp);
	    break;
	}

	/* response to a M_COPYIN message */
	error = 0;
	ret = audio_hlp_iocdata(handle, q, mp, &error);
	switch (ret) {
	case AD_WIOCDATA:
	    /*
	     * caller is responsible for modifying mp to the desired
	     * M_COPYIN/M_COPYOUT message
	     */
	    break;

	default:
	    error = 0;
	case AD_ACK:
	case AD_NACK:
	    iocp = (struct iocblk *)mp->b_rptr;
	    iocp->ioc_count = 0;
	    iocp->ioc_rval = 0;
	    iocp->ioc_error = error;
	    mp->b_datap->db_type = ret == AD_ACK ? M_IOCACK : M_IOCNAK;
	    mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
	    if (mp->b_cont) {
		freemsg(mp->b_cont);
		mp->b_cont = NULL;
	    }
	    break;
	}
	qreply(q, mp);
	break;

    case M_FLUSH:
	/* Canonical flushing code for STREAMS drivers */
	if (*mp->b_rptr & FLUSHW) {
	    flushq(q, FLUSHALL);
	    *mp->b_rptr &= ~FLUSHW;

	    mutex_enter(&handle->lock);

	    if (handle->play_mp) {
		freemsg(handle->play_mp);
		handle->play_mp = 0;
	    }

	    /*
	     * in case the driver is active playing something, stop
	     * the playback, to flush buffered data inside the audio
	     * driver.
	     */
	    if (handle->audio_info.play.active) {
		handle->audio_info.play.active = FALSE;
		mutex_exit(&handle->lock);

		AD_STOP_PLAY(handle, AUDIO_NO_CHANNEL);
	    } else
		mutex_exit(&handle->lock);
	}
	if (*mp->b_rptr & FLUSHR) {
	    flushq(RD(q), FLUSHALL);
	    qreply(q, mp);
	} else {
	    freemsg(mp);
	}
	break;

    case M_DATA:
	if (AUDIOHLP_ISCTL(audio_hlp_get_qptr_dev(q))) {
	    /* don't accept data on audio ctl device */
	    freemsg(mp);
	} else {

	    mutex_enter(&handle->lock);
	    if (!handle->audio_info.play.active) {
		int eof = (mp->b_wptr <= mp->b_rptr);
		if (eof) {
		    /*
		     * an EOF record, does not start audio playback
		     *
		     * reason was the following code sequence:
		     *   fd = open("/dev/audio", O_WRONLY);
		     *   write(fd, 0, 0);
		     *   close(fd);
		     *
		     * Without this special case for EOF records when
		     * playback is not active, it would hang in close,
		     * waiting for the "active" playback to terminate.
		     *
		     * So, we make sure that only a data message
		     * containing real samples starts playback.
		     */
		    handle->audio_info.play.eof++;
		    freemsg(mp);
		    audio_hlp_status_changed(handle);
		} else {
		    /* not an EOF record, start audio */

		    /*
		     * put sample data on queue, for later processing
		     * by audio_hlp_wserv
		     */
		    putq(q, mp);
		    handle->audio_info.play.active = TRUE;
		}
		mutex_exit(&handle->lock);

		if (!eof) 
		    AD_START_PLAY(handle, AUDIO_NO_CHANNEL);
	    } else {
 		/*
		 * put sample data on queue, for later processing by
		 * audio_hlp_wserv
		 */
		putq(q, mp);
		mutex_exit(&handle->lock);
	    }

	}
	break;

    default:
	AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_wput: message type 0x%x unhandled",
		       mp->b_datap->db_type);
	freemsg(mp);
	break;
    }

    return 0;
}


int
audio_hlp_wsvc(queue_t *q)
{
    mblk_t *mp;
    
    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_wsvc\n");

    while ((mp = getq(q)) != NULL) {
	switch (mp->b_datap->db_type) {
	case M_DATA:
	    if (!putbq(q, mp))
		freemsg(mp);

	    /*
	     * It would ne nice if we could push the samples to the
	     * audio driver here, but the audio driver interface
	     * currently has no function for such an operation.
	     */

	    /*
	     * XXX: maybe extend the audio driver interface, like this?
	     *
	    audiohdl_t handle = audio_hlp_get_qptr_data(q);
	    AD_FILL_PLAY_BUFFER(handle, AUDIO_NO_CHANNEL);
	     */

	    return 0;
	default:
	    /*
	     * should not happen, because we only place M_DATA
	     * messages onto our own STREAMS queue in audio_hlp_wput.
	     */
	    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_wserv: message type 0x%x unhandled",
			   mp->b_datap->db_type);
	    freemsg(mp);
	    break;
	}
    }
    return 0;
}


/*
 * Audio Support Module STREAMS Private Data Routines
 */
struct qdata {
    dev_t       dev;
    void       *data;
};


static struct qdata *
audio_hlp_qdata(queue_t *q)
{
    struct qdata *qdata;

    if ((qdata = q->q_ptr) == NULL) {
	qdata = kmem_zalloc(sizeof(struct qdata), KM_SLEEP);
	q->q_ptr = qdata;
	WR(q)->q_ptr = qdata;
    }
    return qdata;
}


void
audio_hlp_set_qptr(queue_t *q, dev_t dev, void *data)
{
    struct qdata *qdata;

    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_set_qptr\n"); */

    qdata = audio_hlp_qdata(q);
    qdata->dev = dev;
    qdata->data = data;
}


void
audio_hlp_free_qptr(queue_t *q)
{				/*  */
    struct qdata *qdata;

    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_free_qptr\n"); */

    if ((qdata = q->q_ptr) == NULL)
	return;

    kmem_free(qdata, sizeof(struct qdata));
    q->q_ptr = NULL;
    WR(q)->q_ptr = NULL;
}


dev_t
audio_hlp_get_qptr_dev(queue_t *q)
{
    struct qdata *qdata;

    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_qptr_dev\n"); */

    if ((qdata = q->q_ptr) == NULL) 
	return NODEV;
    
    return qdata->dev;
}

     
void *
audio_hlp_get_qptr_data(queue_t *q)
{
    struct qdata *qdata;

    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_qptr_data\n"); */

    if ((qdata = q->q_ptr) == NULL) 
	return NULL;

    return qdata->data;
}


int
audio_hlp_get_qptr_instance(queue_t *q)
{
    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_qptr_instance\n"); */

    return AUDIOHLP_INST(audio_hlp_get_qptr_dev(q));
}


/*
 * Audio Support Module Miscellaneous Routines
 */
dev_info_t *
audio_hlp_get_dip(audiohdl_t handle)
{
    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_dip\n"); */

    return handle->ah_dip;
}


void *
audio_hlp_get_private(audiohdl_t handle)
{
    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_private\n"); */

    return handle->ah_private;
}


void
audio_hlp_set_private(audiohdl_t handle, void *private)
{
    /* AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_set_private\n"); */

    handle->ah_private = private;
}


void
audio_hlp_log(audiohdl_t handle, uint_t level, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiohlp_vprintf(level, AUDM_ALL, fmt, args);
    va_end(args);
}


/*
 * Audio Helper Driver Device Dependent Driver Play Routines
 */

/*
 * audio_hlp_get_audio: used by the audio driver to request playback
 * samples from the audio helper framework.
 */
int
audio_hlp_get_audio(audiohdl_t handle, void *buf, int channel, int samples)
{
    mblk_t *mp;
    int src_samples;
    int dst_samples;
    int src_sample_shift;
    int dst_sample_shift;
    audio_prinfo_t *prinfo = &handle->audio_info.play;
    device_format_t *devfmt = &handle->device_play_format;
    int samples_copied;

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_audio %p, samples %d\n",
		   buf, samples);

    if (handle->play_q == NULL) {
	AUD_DPRINTF_L4(AUDM_ALL, "play_q == NULL\n");
	return 0;
    }

    src_sample_shift = 0;
    if (prinfo->precision == 16) src_sample_shift++;
    if (prinfo->channels  ==  2) src_sample_shift++;

    dst_sample_shift = 0;
    if (devfmt->precision == 16) dst_sample_shift++;
    if (devfmt->channels  ==  2) dst_sample_shift++;

    samples_copied = 0;

    if (!mutex_tryenter(&handle->lock)) {
	AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_audio, avoiding deadlock\n");
	return 0;
    }

    while (samples > 0) {
	if ((mp = handle->play_mp) == NULL) {
	    if ((mp = getq(handle->play_q)) == NULL) {
		AUD_DPRINTF_L4(AUDM_ALL, "getq(play_q) == NULL\n");
		break;
	    }
	    if (mp->b_wptr <= mp->b_rptr) {
		/* an EOF audio record; count it */
		prinfo->eof++;
		freemsg(mp);
		audio_hlp_status_changed(handle);
		continue;
	    }
	    handle->play_mp = mp;
	}

	src_samples = (mp->b_wptr - mp->b_rptr) >> src_sample_shift;

	convert_samples(&handle->play_conv_ctx,
			buf, samples, &dst_samples,
			mp->b_rptr, src_samples, &src_samples);

	buf = (char *)buf + (dst_samples << dst_sample_shift);
	mp->b_rptr += src_samples << src_sample_shift;

	if (mp->b_rptr + (1<<src_sample_shift) > mp->b_wptr) {
	    handle->play_mp = mp->b_cont;
	    mp->b_cont = NULL;
	    freemsg(mp);
	}

	prinfo->samples += src_samples;
	samples_copied += dst_samples;
	samples -= dst_samples;
    }

    mutex_exit(&handle->lock);

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_get_audio, returned %d samples\n",
		   samples_copied);

    return samples_copied;
}


/*
 * audio_hlp_play_shutdown: used by the audio driver to signal that
 * playback has stopped due to an underflow condition.
 */
void
audio_hlp_play_shutdown(audiohdl_t handle, int channel)
{
    int start_audio;
    int qs;

    mutex_enter(&handle->lock);
    /*
     * new sample data might got buffered in the playback queue, while
     * the audio driver decided to shutdown playback.  If that is the
     * case, we need to restart playback immediatelly after this
     * shutdown.
     */
    qs = qsize(handle->play_q);
    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_play_shutdown: qsize %d\n", qs);
    handle->audio_info.play.active = start_audio = (qs > 0);
    if (!handle->drain_waiting)
	handle->audio_info.play.error = TRUE;
    cv_signal(&handle->playback_empty_cv);
    /* send a status notification because of a playback underflow condition */
    audio_hlp_status_changed(handle);
    mutex_exit(&handle->lock);

    if (start_audio)
	AD_START_PLAY(handle, AUDIO_NO_CHANNEL);	
}

/*
 * Audio Helper Driver Device Dependent Driver Record Routines
 */
void
audio_hlp_send_audio(audiohdl_t handle, void *buf, int channel, int samples)
{
#define	RECORD_MSG_CHUNK 8192
    mblk_t *mp = NULL;
    int src_sample_shift;
    int dst_sample_shift;
    int src_samples;
    int dst_samples;
    audio_prinfo_t *prinfo = &handle->audio_info.record;
    device_format_t *devfmt = &handle->device_record_format;

    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_send_audio %p %d\n", buf, samples);

    if (buf == NULL || samples == 0 || handle->record_q == NULL) {
	return;
    }

    src_sample_shift = 0;
    if (devfmt->precision == 16) src_sample_shift++;
    if (devfmt->channels  ==  2) src_sample_shift++;

    dst_sample_shift = 0;
    if (prinfo->precision == 16) dst_sample_shift++;
    if (prinfo->channels  ==  2) dst_sample_shift++;

    while (samples > 0) {
	mp = allocb(RECORD_MSG_CHUNK, BPRI_LO);
	if (mp == NULL) {
	    cmn_err(CE_NOTE, "send_audio: no mem for record data");
	    goto error;
	}

	convert_samples(&handle->record_conv_ctx,
			mp->b_rptr, RECORD_MSG_CHUNK >> dst_sample_shift, &dst_samples,
			buf, samples, &src_samples);

	if (src_samples <= 0) {
	    cmn_err(CE_NOTE, "send_audio: convert_samples failed");
	    goto error;
	}
	buf = (char*)buf + (src_samples << src_sample_shift);
	samples -= src_samples;

	if (dst_samples > 0) {
	    mp->b_wptr = mp->b_rptr + (dst_samples << dst_sample_shift);

	    if (!canput(handle->record_q)) {
		AUD_DPRINTF_L3(AUDM_ALL, "send_audio: overflow\n");
		goto error;
	    }

	    if (!putq(handle->record_q, mp)) {
		cmn_err(CE_NOTE, "send_audio: putq failed");
		goto error;
	    }

	    handle->audio_info.record.samples += dst_samples;
	} else {
	    freemsg(mp);
	}
    }

    return;

 error:
    if (mp != NULL)
	freemsg(mp);
    handle->audio_info.record.error = TRUE;
    audio_hlp_status_changed(handle);
    return;
}


/*
 * Audio Mixer Driver Device Dependent Driver Miscellaneous Routines
 */
int audio_hlp_hw_state_change(audiohdl_t handle, int cmd, int dir, int value,
			      int sleep)
{
    AUD_DPRINTF_L4(AUDM_ALL, "audio_hlp_hw_state_change, "
		   "cmd %d, dir %d, val %d, sleep %d\n",
		   cmd, dir, value, sleep);

    /* XXX: implementation this... */

    return AUDIO_SUCCESS;
}
